from typing import Dict, Any
from pos_model import Order
from _NfceXmlPartBuilder import NfceXmlPartBuilder


class EmitBuilder(NfceXmlPartBuilder):
    def __init__(self,
                 crt,
                 cnpj_contribuinte,
                 inscr_estadual,
                 nome_emit,
                 end_logradouro,
                 end_numero,
                 end_compl,
                 bairro,
                 c_mun_fg,
                 municipio,
                 uf,
                 cep):
        # type: (int, unicode, unicode, unicode, unicode, unicode, unicode, unicode, unicode, unicode, unicode, unicode) -> None

        self.crt = crt
        self.cnpj_contribuinte = cnpj_contribuinte
        self.nome_emit = nome_emit
        self.end_logradouro = end_logradouro
        self.end_numero = end_numero
        self.end_compl = end_compl
        self.bairro = bairro
        self.c_mun_fg = c_mun_fg
        self.uf = uf
        self.cep = cep
        self.inscr_estadual = inscr_estadual
        self.municipio = municipio

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        # Dados Emitente
        # CRT: 1 = simples nacional, 2 = simples (excesso de sublimite de recieta), 3 = regime normal
        inner_xml = "<emit>" \
                    "<CNPJ>%s</CNPJ>" \
                    "<xNome>%s</xNome>" \
                    "<enderEmit>" \
                    "<xLgr>%s</xLgr>" \
                    "<nro>%s</nro>" % (self.cnpj_contribuinte, self.nome_emit, self.end_logradouro, self.end_numero)

        if self.end_compl:
            inner_xml += "<xCpl>%s</xCpl>" % self.end_compl

        inner_xml += "<xBairro>%s</xBairro>" \
               "<cMun>%s</cMun>" \
               "<xMun>%s</xMun>" \
               "<UF>%s</UF>" \
               "<CEP>%s</CEP>" \
               "</enderEmit>" \
               "<IE>%s</IE>" \
               "<CRT>%d</CRT>" \
               "</emit>" % (self.bairro, self.c_mun_fg, self.municipio, self.uf, self.cep, self.inscr_estadual, self.crt)

        return inner_xml



from ._StartServiceInteractor import StartServiceInteractor
from ._CreateTableOrderInteractor import CreateTableOrderInteractor
from ._ReopenTableInteractor import ReopenTableInteractor
from ._CloseTableInteractor import CloseTableInteractor
from ._SetTableAvailableInteractor import SetTableAvailableInteractor
from ._DoChangeTipInteractor import DoChangeTipInteractor
from ._ChooseLineSeatInteractor import ChooseLineSeatInteractor
from ._NCouponInteractor import NCouponInteractor
from ._NDiscountInteractor import NDiscountInteractor
from ._NavigationInteractor import NavigationInteractor
from ._ServicePeriodInteractor import ServicePeriodInteractor
from ._OpenDayInteractor import OpenDayInteractor
from ._OrderInteractor import OrderInteractor
from ._DailyGoalsInteractor import DailyGoalsInteractor
from ._MessageDialogInteractor import MessageDialogInteractor

__all__ = [
    "StartServiceInteractor",
    "CreateTableOrderInteractor",
    "ReopenTableInteractor",
    "CloseTableInteractor",
    "SetTableAvailableInteractor",
    "DoChangeTipInteractor",
    "ChooseLineSeatInteractor",
    "NCouponInteractor",
    "NDiscountInteractor",
    "ServicePeriodInteractor",
    "NavigationInteractor",
    "OpenDayInteractor",
    "OrderInteractor",
    "DailyGoalsInteractor",
    "MessageDialogInteractor"
]



import json

from application.interactor import ApplyBenefitInteractor
from application.model import AsyncUuidMessageProcessor
from messagebus import DefaultToken, DataType, Message, MessageBus, Event
from typing import Optional, Any


class ApplyBenefitProcessor(AsyncUuidMessageProcessor):
    
    def __init__(self, interactor):
        # type: (ApplyBenefitInteractor) -> None

        super(ApplyBenefitProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data):
        # type: (str) -> (int, int, str)

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        originator_pos_id = int(parsed_data.get("originatorPosId") or pos_id)
        voucher_id = str(parsed_data.get("voucherId"))

        return pos_id, originator_pos_id, voucher_id

    def call_business(self, params):
        # type: ((int, int, str)) -> None

        pos_id, originator_pos_id, voucher_id = params
        self.interactor.apply_benefit(pos_id, originator_pos_id, voucher_id)

    def format_response(self, message_bus, message, event, request, response):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, Any) -> None
        
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, response))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], Any, BaseException) -> None
        
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        # type: (MessageBus, Optional[Message], Optional[Event], bytes, BaseException) -> None
        
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



# -*- coding: utf-8 -*-

import os
from datetime import datetime, timedelta
from xml.etree import cElementTree as eTree

import iso8601
from helper import BaseRepository, ReportType
from msgbus import MBEasyContext  # noqa
from persistence import Connection  # noqa
from systools import sys_log_exception
from typing import List, Dict, Optional  # noqa

from reports_app.cashreport.dto import Order, OrderTender


class OrderRepository(BaseRepository):
    def __init__(self, mbcontext, pos_list, fiscal_sent_dir):
        # type: (MBEasyContext, List[int], Optional[unicode]) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list
        self.fiscal_sent_dir = fiscal_sent_dir

    def get_paid_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_paid_orders(ReportType.RealDate, initial_date, end_date, operator_id, None, report_pos)

    def get_paid_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_paid_orders(ReportType.BusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_paid_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_paid_orders(ReportType.SessionId, None, None, None, session_id, None)

    def get_voided_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(ReportType.RealDate, initial_date, end_date, operator_id, None, report_pos)

    def get_voided_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_voided_orders(ReportType.BusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_voided_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_voided_orders(ReportType.SessionId, None, None, None, session_id, None)

    def get_paid_orders_by_xml(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Order]
        orders = []  # type: List[Order]

        current_date = initial_date
        while current_date <= end_date:
            xml_directory = os.path.join(self.fiscal_sent_dir, "Enviados", current_date.strftime("%Y"), current_date.strftime("%m"), current_date.strftime("%d"))
            if os.path.exists(xml_directory):
                for entry in os.listdir(xml_directory):
                    try:
                        if entry.endswith(".xml"):
                            with open(os.path.join(xml_directory, entry), "rb") as xml_file:
                                xml_content = xml_file.read()
                            order = self._get_order_from_xml(entry, xml_content)
                            if order is not None:
                                orders.append(order)
                    except Exception as ex:
                        sys_log_exception("Fail to read xml: %s; Exception: %s" % (xml_directory, ex))

            current_date += timedelta(days=1)

        return orders

    def _get_paid_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]
            try:
                if report_type == ReportType.RealDate:
                    query = self._PaidOrdersByRealDateQuery
                    date_format = "%Y-%m-%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                elif report_type == ReportType.BusinessPeriod:
                    query = self._PaidOrderByBusinessPeriod
                    date_format = "%Y%m%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                else:
                    query = self._PaidOrderBySessionId
                    date_format = "%Y%m%d"
                    query = query.format(session_id)

                if (report_type == ReportType.RealDate or ReportType.BusinessPeriod) and operator_id is not None:
                    query += " and o.SessionId like '%user={0},%'".format(operator_id)

                all_orders = {}  # type: Dict[int, Order]
                order_with_payments = [(int(x.get_entry(0)),
                                        int(x.get_entry(1)),
                                        datetime.strptime(x.get_entry(2), date_format),
                                        float(x.get_entry(3)),
                                        int(x.get_entry(4)),
                                        float(x.get_entry(5)),
                                        float(x.get_entry(6)),
                                        float(x.get_entry(7))) for x in conn.select(query)]

                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_state = order_tuple[1]
                    order_date = order_tuple[2]
                    order_total = order_tuple[3]
                    donation_value = order_tuple[6]
                    change_amount = order_tuple[7]

                    if order_id in all_orders:
                        order = all_orders[order_id]
                    else:
                        order = Order(order_id, order_date, order_state, order_total, donation_value, [], None, change_amount)
                        all_orders[order_id] = order

                    tender_type = order_tuple[4]
                    tender_value = order_tuple[5]

                    order_tender = OrderTender(tender_type, tender_value)
                    order.tenders.append(order_tender)

                return list(all_orders.values())
            except:
                return list()

        report_pos_list = self.pos_list if report_pos is None else (report_pos, )
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    def _get_voided_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]
            if report_type == ReportType.RealDate:
                query = self._VoidedOrdersByRealDateQuery
                date_format = "%Y-%m-%d"
                query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
            elif report_type == ReportType.BusinessPeriod:
                query = self._VoidedOrdersByBusinessPeriod
                date_format = "%Y%m%d"
                query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
            else:
                query = self._VoidedOrdersBySessionId
                date_format = "%Y%m%d"
                query = query.format(session_id)

            if (report_type == ReportType.RealDate or report_type == ReportType.BusinessPeriod) and operator_id is not None:
                query += " and o.SessionId like '%user={0},%'".format(operator_id)

            order_with_payments = [(int(x.get_entry(0)),
                                    int(x.get_entry(1)),
                                    datetime.strptime(x.get_entry(2), date_format),
                                    float(x.get_entry(3)),
                                    float(x.get_entry(4)),
                                    x.get_entry(5)) for x in conn.select(query)]

            all_orders = []  # type: List[Order]
            for order_tuple in order_with_payments:
                order_id = order_tuple[0]
                order_state = order_tuple[1]
                order_date = order_tuple[2]
                order_total = order_tuple[3]
                donation_value = order_tuple[4]
                void_reason_id = order_tuple[5]

                order = Order(order_id, order_date, order_state, order_total, donation_value, [], void_reason_id, None)
                all_orders.append(order)

            return all_orders

        report_pos_list = self.pos_list if report_pos is None else (report_pos, )
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    def _get_order_from_xml(self, xml_file_name, xml_content):
        # type: (str, str) -> Optional[Order]
        if xml_content.find("<CFe>") >= 0:
            return self._get_order_from_sat_xml(xml_file_name, xml_content)
        elif xml_content.find("<NFe") >= 0:
            return self._get_order_from_nfce_xml(xml_content)
        else:
            return None

    @staticmethod
    def _get_order_from_sat_xml(xml_file_name, xml_content):
        parsed_xml = eTree.XML(xml_content)
        index1 = xml_file_name.find("_")
        index2 = xml_file_name.find("_", index1 + 1)
        order_id = int(xml_file_name[index1 + 1: index2])

        numero_nota = xml_file_name[:index1]

        data_nota = parsed_xml.find("infCFe/ide/dEmi").text
        hora_nota = parsed_xml.find("infCFe/ide/hEmi").text
        data_emissao = datetime.strptime(data_nota + hora_nota, "%Y%m%d%H%M%S")

        v_total = float(parsed_xml.find("infCFe/total/ICMSTot/vProd").text)

        tenders = []  # type: List[OrderTender]
        troco = parsed_xml.find("infCFe/pgto/vTroco")
        pgtos = parsed_xml.findall("infCFe/pgto/MP")
        for pgto in pgtos:
            c_mp = int(pgto.find("cMP").text)
            v_mp = float(pgto.find("vMP").text)

            if c_mp == 1:  # DINHEIRO
                tender_type = 0
            elif c_mp == 3:  # CREDITO:
                tender_type = 1
            elif c_mp == 4:  # DEBITO
                tender_type = 2
            else:
                tender_type = 99

            if tender_type == 0 and troco is not None:
                v_mp -= float(troco.text)
                troco = None

            tender = OrderTender(tender_type, v_mp)
            tenders.append(tender)

        order = Order(order_id, data_emissao, 5, v_total, 0, tenders, None)
        order.numero_nota = numero_nota

        return order

    @staticmethod
    def _get_order_from_nfce_xml(xml_content):
        nfe_namespace = "http://www.portalfiscal.inf.br/nfe"

        parsed_xml = eTree.XML(xml_content)
        c_nf = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}cNF".format(nfe_namespace)).text
        n_serie = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}serie".format(nfe_namespace)).text
        order_id = int(n_serie + c_nf)

        data_hora_nota = parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(nfe_namespace)).text
        data_emissao = iso8601.parse_date(data_hora_nota)

        v_total = float(parsed_xml.find("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}total/{{{0}}}ICMSTot/{{{0}}}vProd".format(nfe_namespace)).text)

        tenders = []  # type: List[OrderTender]
        pgtos = parsed_xml.findall("{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}pag".format(nfe_namespace))
        for pgto in pgtos:
            t_pag = int(pgto.find("{{{0}}}tPag".format(nfe_namespace)).text)
            v_pag = float(pgto.find("{{{0}}}vPag".format(nfe_namespace)).text)

            if t_pag == 1:
                tender_type = 0  # DINHEIRO
            elif t_pag == 3:
                tender_type = 1  # CREDITO
            elif t_pag == 4:
                tender_type = 2  # DEBITO
            else:
                tender_type = 99  # OUTROS

            tender = OrderTender(tender_type, v_pag)
            tenders.append(tender)

        order = Order(order_id, data_emissao, 5, v_total, 0, tenders, None)

        return order

    _PaidOrdersByRealDateQuery = \
"""select o.OrderId, o.StateId, od.OrderDate, o.TotalGross, ot.TenderId, ot.TenderAmount, coalesce(ocp.Value, 0), coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
inner join
(
    -- Busca as Orders do dia que estamos interessadosh
    select OrderId, strftime('%Y-%m-%d', Value, 'localtime') OrderDate
    from OrderCustomProperties
    where key = 'FISCALIZATION_DATE' and strftime('%Y-%m-%d', Value, 'localtime') >= '{0}' and strftime('%Y-%m-%d', Value, 'localtime') <= '{1}'
) od
on o.OrderId = od.OrderId
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
where o.StateId = 5;"""

    _VoidedOrdersByRealDateQuery = \
"""select o.OrderId, o.StateId, od.OrderDate, coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value as VoidReason
from Orders o
inner join
(
    -- Busca as Orders do dia que estamos interessados
    select OrderId, OrderDate
    from
    (
        -- Busca a última data que as Orders foram totalizadas
        select OrderId, strftime('%Y-%m-%d', max(Timestamp) , 'localtime') OrderDate
        from
        (
            select OrderId, StateId, Timestamp from OrderStateHistory
            where strftime('%Y-%m-%d', Timestamp, 'localtime') > strftime('%Y-%m-%d', '{0}' , '-1 day') and strftime('%Y-%m-%d', Timestamp, 'localtime') < strftime('%Y-%m-%d', '{1}' , '+1 day')
        )
        where StateId = 4
        group by OrderId
    ) a where a.OrderDate >= '{0}' and a.OrderDate <= '{1}'
) od
on o.OrderId = od.OrderId
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
where o.StateId = 4;"""

    _PaidOrderByBusinessPeriod = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, ot.TenderAmount, coalesce(ocp.Value, 0), coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
where o.BusinessPeriod >= '{0}' and o.BusinessPeriod <= '{1}' and o.StateId = 5;"""

    _VoidedOrdersByBusinessPeriod = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value as VoidReason
from Orders o
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
where o.BusinessPeriod >= '{0}'  and o.BusinessPeriod <= '{1}' and o.StateId = 4;"""

    _PaidOrderBySessionId = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, ot.TenderAmount, coalesce(ocp.Value, 0), coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
where o.SessionId = '{0}' and o.StateId = 5;"""

    _VoidedOrdersBySessionId = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, coalesce(o.TotalGross, 0), coalesce(ocp.Value, 0), vrs.Value as VoidReason
from Orders o
left join OrderCustomProperties ocp
on ocp.OrderId = o.OrderId and ocp.Key = 'DONATION_VALUE'
left join OrderCustomProperties vrs on vrs.OrderId = o.OrderId and vrs.Key = 'VOID_REASON_ID'
where o.SessionId = '{0}' and o.StateId = 4;"""



import glob
from datetime import datetime, timedelta
from os import remove, path


class CleanJsonProcessors(object):
    def __init__(self, json_paths, keep_days):
        self.json_paths = json_paths
        self.validation_date = datetime.now() - timedelta(days=keep_days)

    def clean_json(self):
        for json_path in self.json_paths:
            json_files_in_path = glob.glob("{}/*.json".format(json_path))
            for json_file in json_files_in_path:
                if datetime.fromtimestamp(path.getmtime(json_file)) < self.validation_date:
                    remove(json_file)



# -*- coding: utf-8 -*-
"""The code for async support. Importing this patches Jinja on supported
Python versions.
"""
import asyncio
import inspect
from functools import update_wrapper

from markupsafe import Markup

from .environment import TemplateModule
from .runtime import LoopContext
from .utils import concat
from .utils import internalcode
from .utils import missing


async def concat_async(async_gen):
    rv = []

    async def collect():
        async for event in async_gen:
            rv.append(event)

    await collect()
    return concat(rv)


async def generate_async(self, *args, **kwargs):
    vars = dict(*args, **kwargs)
    try:
        async for event in self.root_render_func(self.new_context(vars)):
            yield event
    except Exception:
        yield self.environment.handle_exception()


def wrap_generate_func(original_generate):
    def _convert_generator(self, loop, args, kwargs):
        async_gen = self.generate_async(*args, **kwargs)
        try:
            while 1:
                yield loop.run_until_complete(async_gen.__anext__())
        except StopAsyncIteration:
            pass

    def generate(self, *args, **kwargs):
        if not self.environment.is_async:
            return original_generate(self, *args, **kwargs)
        return _convert_generator(self, asyncio.get_event_loop(), args, kwargs)

    return update_wrapper(generate, original_generate)


async def render_async(self, *args, **kwargs):
    if not self.environment.is_async:
        raise RuntimeError("The environment was not created with async mode enabled.")

    vars = dict(*args, **kwargs)
    ctx = self.new_context(vars)

    try:
        return await concat_async(self.root_render_func(ctx))
    except Exception:
        return self.environment.handle_exception()


def wrap_render_func(original_render):
    def render(self, *args, **kwargs):
        if not self.environment.is_async:
            return original_render(self, *args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.render_async(*args, **kwargs))

    return update_wrapper(render, original_render)


def wrap_block_reference_call(original_call):
    @internalcode
    async def async_call(self):
        rv = await concat_async(self._stack[self._depth](self._context))
        if self._context.eval_ctx.autoescape:
            rv = Markup(rv)
        return rv

    @internalcode
    def __call__(self):
        if not self._context.environment.is_async:
            return original_call(self)
        return async_call(self)

    return update_wrapper(__call__, original_call)


def wrap_macro_invoke(original_invoke):
    @internalcode
    async def async_invoke(self, arguments, autoescape):
        rv = await self._func(*arguments)
        if autoescape:
            rv = Markup(rv)
        return rv

    @internalcode
    def _invoke(self, arguments, autoescape):
        if not self._environment.is_async:
            return original_invoke(self, arguments, autoescape)
        return async_invoke(self, arguments, autoescape)

    return update_wrapper(_invoke, original_invoke)


@internalcode
async def get_default_module_async(self):
    if self._module is not None:
        return self._module
    self._module = rv = await self.make_module_async()
    return rv


def wrap_default_module(original_default_module):
    @internalcode
    def _get_default_module(self):
        if self.environment.is_async:
            raise RuntimeError("Template module attribute is unavailable in async mode")
        return original_default_module(self)

    return _get_default_module


async def make_module_async(self, vars=None, shared=False, locals=None):
    context = self.new_context(vars, shared, locals)
    body_stream = []
    async for item in self.root_render_func(context):
        body_stream.append(item)
    return TemplateModule(self, context, body_stream)


def patch_template():
    from . import Template

    Template.generate = wrap_generate_func(Template.generate)
    Template.generate_async = update_wrapper(generate_async, Template.generate_async)
    Template.render_async = update_wrapper(render_async, Template.render_async)
    Template.render = wrap_render_func(Template.render)
    Template._get_default_module = wrap_default_module(Template._get_default_module)
    Template._get_default_module_async = get_default_module_async
    Template.make_module_async = update_wrapper(
        make_module_async, Template.make_module_async
    )


def patch_runtime():
    from .runtime import BlockReference, Macro

    BlockReference.__call__ = wrap_block_reference_call(BlockReference.__call__)
    Macro._invoke = wrap_macro_invoke(Macro._invoke)


def patch_filters():
    from .filters import FILTERS
    from .asyncfilters import ASYNC_FILTERS

    FILTERS.update(ASYNC_FILTERS)


def patch_all():
    patch_template()
    patch_runtime()
    patch_filters()


async def auto_await(value):
    if inspect.isawaitable(value):
        return await value
    return value


async def auto_aiter(iterable):
    if hasattr(iterable, "__aiter__"):
        async for item in iterable:
            yield item
        return
    for item in iterable:
        yield item


class AsyncLoopContext(LoopContext):
    _to_iterator = staticmethod(auto_aiter)

    @property
    async def length(self):
        if self._length is not None:
            return self._length

        try:
            self._length = len(self._iterable)
        except TypeError:
            iterable = [x async for x in self._iterator]
            self._iterator = self._to_iterator(iterable)
            self._length = len(iterable) + self.index + (self._after is not missing)

        return self._length

    @property
    async def revindex0(self):
        return await self.length - self.index

    @property
    async def revindex(self):
        return await self.length - self.index0

    async def _peek_next(self):
        if self._after is not missing:
            return self._after

        try:
            self._after = await self._iterator.__anext__()
        except StopAsyncIteration:
            self._after = missing

        return self._after

    @property
    async def last(self):
        return await self._peek_next() is missing

    @property
    async def nextitem(self):
        rv = await self._peek_next()

        if rv is missing:
            return self._undefined("there is no next item")

        return rv

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._after is not missing:
            rv = self._after
            self._after = missing
        else:
            rv = await self._iterator.__anext__()

        self.index0 += 1
        self._before = self._current
        self._current = rv
        return rv, self


async def make_async_loop_context(iterable, undefined, recurse=None, depth0=0):
    import warnings

    warnings.warn(
        "This template must be recompiled with at least Jinja 2.11, or"
        " it will fail in 3.0.",
        DeprecationWarning,
        stacklevel=2,
    )
    return AsyncLoopContext(iterable, undefined, recurse, depth0)


patch_all()



# -*- coding: utf-8 -*-
import base64
import logging
import os
import time
import clr

logger = logging.getLogger("Cappta")


class CapptaProcessor(object):
    RESPONDE_XML = """<EftXml Media='%s' Result='%s' CardNumber='%sXXXXXXXXXX' AuthCode='%s' ApprovedAmount='%s' IdAuth='%s' CNPJAuth='%s' Bandeira='%s' ReceiptMerchant='%s' ReceiptCustomer='%s' TransactionProcessor='%s'/>"""

    def process(self, bus_msg, pos_id, order_id, operator, tender_type, amount, fiscal_date, fiscal_time, tender_id):
        raise NotImplementedError()

    def finalizar(self, pos_id, order_id, data_fiscal, hora_fiscal, status):
        return

    def reprint(self, bus_msg, pos_id, slip_type):
        raise NotImplementedError()

    def cancel(self, bus_msg, pos_id, password, control_number):
        raise NotImplementedError()


class FakeCapptaProcessor(CapptaProcessor):
    def __init__(self):
        super(FakeCapptaProcessor, self).__init__()

    def process(self, bus_msg, pos_id, order_id, operator, tender_type, amount, fiscal_date, fiscal_time, tender_id):
        nome_instituicao = ''
        ret = '0'
        bin_cartao = '4111XXXXXXXXXXXX'
        auth_code = '1302'
        rede_autorizadora = '08496898000142'
        bandeira = '4'
        transaction_processor = '00051'
        b64_receipt_customer = base64.b64encode("Recibo de teste.\nSitef FAKE")
        b64_receipt_merchant = base64.b64encode("Recibo de teste.\nSitef FAKE")

        response = CapptaProcessor.RESPONDE_XML % (nome_instituicao, ret, bin_cartao, auth_code, amount, rede_autorizadora, "", bandeira, b64_receipt_merchant, b64_receipt_customer, transaction_processor)
        return response

    def finalizar(self, pos_id, order_id, data_fiscal, hora_fiscal, status):
        return

    def reprint(self, bus_msg, pos_id, slip_type):
        return

    def cancel(self, bus_msg, pos_id, password, control_number):
        return


class SynchCapptaProcessor(CapptaProcessor):
    def __init__(self, pos_id, pdv_tef, cnpj, auth_key, module_directory, callback, env):
        super(SynchCapptaProcessor, self).__init__()
        self.cnpj = cnpj
        self.pdv_tef = pdv_tef
        self.pos_id = int(pos_id)
        self.auth_key = auth_key
        self.env = env
        self.callback = callback
        self.retorno = ["Sucesso",
                        "Não autenticado. Alguma das informações fornecidas para autenticação não é válida",
                        "CapptaGpPlus está sendo inicializado",
                        "Formato da requisição recebida pelo CapptaGpPlus é inválido",
                        "Operação cancelada pelo operador",
                        "Pagamento não autorizado/pendente/não encontrado",
                        "Pagamento ou cancelamento negados pela rede adquirente ou falta de conexão com internet",
                        "Erro interno no CapptaGpPlus",
                        "Erro na comunicação entre a CappAPI e o CapptaGpPlus",
                        "Último pagamento ainda não finalizado",
                        "Uma reimpressão ou cancelamento foi executada dentro de uma sessão multi-cartões",
                        "", "", "",
                        "Valor digitado no pinpad é inválido"]
        old_dir = os.getcwd()
        os.chdir(os.path.join(module_directory, "dll"))
        clr.AddReference(os.path.join(os.getcwd(), "Cappta.Gp.Api.Com.dll"))

        from Cappta.Gp.Api.Com import ClienteCappta
        self.clienteCappta = ClienteCappta()
        os.chdir(old_dir)

        if self.env == 2:
            logger.info(">>>>> Ambiente de homologação ativado! <<<<<")

    def process(self, bus_msg, pos_id, order_id, operator, tender_type, amount, fiscal_date, fiscal_time, tender_id):
        from Cappta.Gp.Api.Com.Model import DetalhesCredito, IMensagem, IRespostaTransacaoPendente, \
            IRequisicaoParametro, IRespostaOperacaoAprovada, IRespostaOperacaoRecusada

        starting_message = "POS {} - Iniciando pagamento para order {}"
        logger.debug(starting_message.format(str(self.pos_id if self.env == 1 else 2), order_id))
        ret = self.clienteCappta.AutenticarPdv(self.cnpj, self.pdv_tef if self.env == 1 else 2, self.auth_key)
        if ret == 0:
            logger.debug("POS %s - Autenticação com sucesso." % (str(self.pos_id if self.env == 1 else 2)))
            self.clienteCappta.IniciarMultiCartoes(9)
            tender_type = int(tender_type)

            if tender_type == 2:  # Debito
                ret = self.clienteCappta.PagamentoDebito(float(amount))
            elif tender_type == 1:  # Credito
                details = DetalhesCredito()
                ret = self.clienteCappta.PagamentoCredito(float(amount), details)
            if ret != 0:
                self.callback(bus_msg, pos_id, ret, "")
                return

            tef_loop_variable = self.clienteCappta.IterarOperacaoTef()
            while True:
                if isinstance(tef_loop_variable, IMensagem):
                    message = IMensagem(tef_loop_variable)
                    msg = message.Descricao
                    self.callback(message, str(self.pos_id), msg, "MENSAGEM")
                    time.sleep(0.5)

                elif isinstance(tef_loop_variable, IRespostaTransacaoPendente):
                    pending_transactions = IRespostaTransacaoPendente(tef_loop_variable)
                    msg = pending_transactions.Mensagem
                    parameter = self.callback(bus_msg, pos_id, str(msg.encode('ascii', 'ignore')), "TRANSACAO_PENDENTE")
                    ret = self.clienteCappta.EnviarParametro(parameter, 1)

                elif isinstance(tef_loop_variable, IRequisicaoParametro):
                    req_parameter = IRequisicaoParametro(tef_loop_variable)
                    msg = req_parameter.Mensagem.encode('utf-8')
                    if "Digite" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_NUMERO_CARTAO")
                    elif "Validade" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_VALIDADE")
                    elif "desconecte" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_DESCONECTAR_PINPAD")
                    elif "cancelar" or "Cancelada" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_ACAO")
                    elif "Seguranca" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_CODIGO_SEGURANÇA")
                    elif "indisponível" in msg:
                        parameter = self.callback(bus_msg, str(self.pos_id), msg, "REQUISICAO_ACAO")
                    else:
                        parameter = 0
                    if parameter == "CANCELA":
                        ret = self.clienteCappta.EnviarParametro("", 2)
                    else:
                        ret = self.clienteCappta.EnviarParametro(parameter, 1)

                if isinstance(tef_loop_variable, IRespostaOperacaoAprovada) or isinstance(tef_loop_variable, IRespostaOperacaoRecusada):
                    break

                tef_loop_variable = self.clienteCappta.IterarOperacaoTef()

            if isinstance(tef_loop_variable, IRespostaOperacaoAprovada):
                approved_payment = IRespostaOperacaoAprovada(tef_loop_variable)
                client_slip = approved_payment.CupomCliente
                customer_slip = approved_payment.CupomLojista
                auth_code = str(approved_payment.CodigoAutorizacaoAdquirente)
                nome_instituicao = str(approved_payment.NomeBandeiraCartao)
                card_bin = str(approved_payment.NumeroCartaoCliente)[:6]
                rede_autorizadora = ""
                transaction_processor = ""
                bandeira = str(approved_payment.CodigoBandeiraCartao)

                b64_receipt_merchant = base64.b64encode(str(customer_slip).replace("\"", ""))
                b64_receipt_customer = base64.b64encode(str(client_slip).replace("\"", ""))
                response = CapptaProcessor.RESPONDE_XML % (nome_instituicao, ret, card_bin, auth_code, amount, rede_autorizadora, "", bandeira, b64_receipt_merchant, b64_receipt_customer, transaction_processor)
                logger.debug("POS %s - Operação aprovada. Order %s" % (str(self.pos_id), str(order_id)))
                return True, response

            elif isinstance(tef_loop_variable, IRespostaOperacaoRecusada):
                operacao = IRespostaOperacaoRecusada(tef_loop_variable)
                response = operacao.Motivo.encode('utf-8')
                logger.debug("POS %s - Operação recusada. Order: %s. Motivo: %s" % (str(self.pos_id), str(order_id), response))
                response = "<EftXml Result=\"%s\" />" % response
                return True, response
        else:
            logger.error("Falha ao autenticar pdv %s - Código de erro %s" % (str(self.pos_id), str(ret)))
            response = "%s" % self.retorno[ret] + ". Código de erro %s" % str(ret)
            response = "<EftXml Result=\"%s\" />" % response
            return True, response

    def finalizar(self, pos_id, order_id, data_fiscal, hora_fiscal, status):
        if status == 1:
            self.clienteCappta.ConfirmarPagamentos()
            logger.debug("POS %s - Pagamentos confirmados. Order: %s" % (str(self.pos_id), str(order_id)))
        else:
            self.clienteCappta.DesfazerPagamentos()
            logger.debug("POS %s - Pagamentos cancelados. Order: %s" % (str(self.pos_id), str(order_id)))

    def reprint(self, bus_msg, pos_id, slip_type):
        # TODO: Implement CAPPTA reprint
        return

    def cancel(self, bus_msg, pos_id, password, control_number):
        # TODO: Implement CAPPTA cancel
        return



from pos_model import Order, XmlTenderTypeFromPOSMapper

from _SatXmlPartBuilder import SatXmlPartBuilder
from satbuilder import ContextKeys


class CfePagType(object):
    dinheiro = 1
    credito = 3
    debito = 4
    outros = 99


class PgtoBuilder(SatXmlPartBuilder):
    def build_xml(self, order, context):
        # type: (Order, ContextKeys) -> unicode

        xml = "<pgto>"
        # Pagamentos

        template = "<MP><cMP>{0:>02}</cMP><vMP>{1:>0.2f}</vMP><cAdmC>999</cAdmC></MP>"
        if len(order.tenders) <= 10:
            for tender in order.tenders:
                tipo = XmlTenderTypeFromPOSMapper.get(tender.type)
                xml += template.format(tipo, tender.amount)
        else:
            tenders_by_type = {}
            for tender in order.tenders:
                tipo = XmlTenderTypeFromPOSMapper.get(tender.type)
                if tipo not in tenders_by_type:
                    tenders_by_type[tipo] = 0
                tenders_by_type[tipo] += tender.amount

            for tender_type in tenders_by_type:
                xml += template.format(tender_type, tenders_by_type[tender_type])

        xml += "</pgto>"

        return xml



# encoding: utf-8

import json
import xml.etree.cElementTree as eTree

import sysactions
from typing import List, Any, Dict

from repository import SaleChallengeRepository, ProductRepository, OrderRepository
from .model import ProductSaleItem, OperatorSaleItem, SaleChallengeItem
from utils import get_store_id


class SaleChallengeReport(object):

    def __init__(self, sale_challenge_repository, product_repository, order_repository, operator_id):
        # type: (SaleChallengeRepository, ProductRepository, OrderRepository, str) -> None

        self.sale_challenge_repository = sale_challenge_repository
        self.product_repository = product_repository
        self.order_repository = order_repository
        self.store_id = get_store_id()
        self.operator_id = operator_id

    def get_operator_name(self):
        if self.operator_id and self.operator_id not in ["0", "", "None"]:
            return eTree.XML(sysactions.get_user_information(self.operator_id)).find(".//user").get("LongName")
        return None

    def get_sale_challenge(self, sale_challenge_id):
        # type: (int) -> SaleChallengeItem
        return self.sale_challenge_repository.get_sale_challenge(sale_challenge_id)

    def get_sales_items(self, sale_challenge):
        # type:(SaleChallengeItem) -> Any[List[OperatorSaleItem], List[ProductSaleItem]]

        if self.operator_id:
            return self._get_operator_sold_items(sale_challenge)
        return self._get_order_items_all_operators(sale_challenge)

    @staticmethod
    def _get_operators_names():
        # type: () -> Dict[str, str]

        operator_name = {}
        for user in eTree.XML(sysactions.get_user_information()).findall(".//user"):
            operator_name[user.get("UserId")] = user.get("LongName")
        return operator_name

    def _get_order_items_all_operators(self, sale_challenge):
        # type: (SaleChallengeItem) -> List[OperatorSaleItem]

        paid_orders = self.order_repository.get_paid_orders_by_business_period(sale_challenge.start_date,
                                                                               sale_challenge.end_date)
        order_items = {}
        for i, p in enumerate(paid_orders):
            if p:
                self.order_repository.get_quantity_items_by_operator(p, order_items, i)

        operators = self._get_operators_names()

        operator_sales = {}
        for operator_id, operator_name in operators.items():
            if operator_id not in order_items.keys():
                continue

            for product_code, product_quantity in order_items[operator_id].items():
                if int(product_code) not in sale_challenge.products:
                    continue

                if operator_id not in operator_sales:
                    operator_sales[operator_id] = OperatorSaleItem(operator_name, product_quantity)
                else:
                    operator_sales[operator_id].quantity += product_quantity

        return list(operator_sales.values())

    def _get_operator_sold_items(self, sale_challenge):
        # type: (SaleChallengeItem) -> List[ProductSaleItem]

        product_names = self.product_repository.get_items_names(sale_challenge.products)
        initial_date = sale_challenge.start_date.strftime("%Y-%m-%d %H:%M:%S.0")
        end_date = sale_challenge.end_date.strftime("%Y-%m-%d %H:%M:%S.0")

        product_items = json.loads(sysactions.generate_report("pmix_order_items", initial_date, end_date,
                                                        self.store_id, self.operator_id))

        sold_items = {}
        for order_item in product_items:
            product_code = order_item['pcode']
            if product_code not in sale_challenge.products:
                continue

            if product_code not in sold_items:
                product_name = product_names[product_code]
                sold_items[product_code] = ProductSaleItem(product_name, order_item['quantity'],
                                                      order_item['quantity'] * float(order_item["unit_price"]))
            else:
                sold_items[product_code].quantity += order_item['quantity']
                sold_items[product_code].total_price += order_item['quantity'] * float(order_item["unit_price"])

        return list(sold_items.values())



from abc import ABCMeta, abstractmethod
from mw_helper import BaseRepository


class CurrencyConversionRepository(BaseRepository):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_currency_exchange(self):
        # type: () -> float
        raise NotImplementedError()



# -*- coding: utf-8 -*-

from sysactions import action, show_messagebox # noqa
from bustoken import TK_POS_UPDATER_PERFORM_BINARIES_UPDATE
from msgbus import TK_SYS_NAK # noqa

from .. import mb_context


@action
def update_binaries(pos_id):

    resp = show_messagebox(pos_id, "$UPDATE_BINARIES_MESSAGE_QUESTION", buttons="$OK|$CANCEL")
    if resp == 1:
        return

    timeout = 60000000 * 20
    binary_update_token = TK_POS_UPDATER_PERFORM_BINARIES_UPDATE
    msg = mb_context.MB_EasySendMessage("PosUpdater", binary_update_token, timeout=timeout)
    if msg.token == TK_SYS_NAK:
        show_messagebox(pos_id, "$UPDATE_BINARIES_" + msg.data.upper(), buttons="OK")
        return

    show_messagebox(pos_id, "$UPDATE_BINARIES_MESSAGE", buttons="OK")



from ._Command import Command
from ._AlignCommand import AlignCommand
from ._BoldCommand import BoldCommand
from ._RepeatCommand import RepeatCommand
from ._FontCommand import FontCommand
from ._NewLineCommand import NewLineCommand


__all__ = [
    "Command",
    "AlignCommand",
    "BoldCommand",
    "RepeatCommand",
    "FontCommand",
    "NewLineCommand"
]



from helper import ExtendedEnum
from messagebus import TokenCreator, TokenPriority

PERIPHERALS_GROUP = "6"


class ListenedQueueTokens(ExtendedEnum):
    TK_PRN_PRINT = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "101")  # 4032823553
    TK_PRN_GETSTATUS = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "102")
    TK_CDRAWER_OPEN = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "201")
    TK_CDRAWER_GETSTATUS = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "202")



# encoding: utf-8

from old_helper import BaseRepository


class DeliveryProductRepository(BaseRepository):

    def __init__(self, mb_context):
        super(DeliveryProductRepository, self).__init__(mb_context)

    def get_products(self):
        def inner_func(conn):
            products = {}
            query = self._ProductsQuery

            for x in conn.select(query):
                product_code = x.get_entry("ProductCode")
                product_name = x.get_entry("ProductName")
                product_type = x.get_entry("ProductType").replace("TYPE_", "")
                products[product_code] = (product_name, product_type)

            return products

        return self.execute_with_connection(inner_func)

    def get_cfh_products(self):
        def inner_func(conn):
            cfh_products = []
            query = self._CFHQuery

            for x in conn.select(query):
                cfh_products.append(x.get_entry("ProductCode"))

            return cfh_products

        return self.execute_with_connection(inner_func)

    _ProductsQuery = """
                    SELECT DISTINCT p.ProductCode, p.ProductName, pt.TypeDescr as ProductType
                    FROM Product p
                    INNER JOIN ProductKernelParams pkp ON pkp.ProductCode = p.ProductCode
                    JOIN ProductType pt ON pt.TypeId = pkp.ProductType
                  """

    _CFHQuery = """
                    SELECT ProductCode FROM ProductTags
                    WHERE LOWER(Tag) = 'cfh=true'
                    UNION
                    SELECT ProductCode FROM ProductCustomParams
                    WHERE LOWER(CustomParamId) = 'cfh' AND LOWER(CustomParamValue) = 'true'
                """



# -*- coding: utf-8 -*-
import time
import datetime
import calendar


def get_system_offset():
    """Get system's timezone offset using built-in library time.

    For the Timezone constants (altzone, daylight, timezone, and tzname), the
    value is determined by the timezone rules in effect at module load time or
    the last time tzset() is called and may be incorrect for times in the past.

    To keep compatibility with Windows, we're always importing time module here.
    """

    localtime = calendar.timegm(time.localtime())
    gmtime = calendar.timegm(time.gmtime())
    offset = gmtime - localtime
    # We could get the localtime and gmtime on either side of a second switch
    # so we check that the difference is less than one minute, because nobody
    # has that small DST differences.
    if abs(offset - time.altzone) < 60:
        return -time.altzone
    else:
        return -time.timezone


def get_tz_offset(tz):
    """Get timezone's offset using built-in function datetime.utcoffset()."""
    return int(datetime.datetime.now(tz).utcoffset().total_seconds())


def assert_tz_offset(tz):
    """Assert that system's timezone offset equals to the timezone offset found.

    If they don't match, we probably have a misconfiguration, for example, an
    incorrect timezone set in /etc/timezone file in systemd distributions."""
    tz_offset = get_tz_offset(tz)
    system_offset = get_system_offset()
    if tz_offset != system_offset:
        msg = ('Timezone offset does not match system offset: {0} != {1}. '
               'Please, check your config files.').format(
                   tz_offset, system_offset
               )
        raise ValueError(msg)



# -*- coding: utf-8 -*-

from datetime import datetime

from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from systools import sys_log_error
from typing import List, Dict


class OrderRepository(BaseRepository):

    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, List[int]) -> None

        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_quantity_items_by_operator(self, all_orders, all_order_items, pos_index):
        # type: (Dict[str, List], Dict[str, Dict], int) -> Dict[str, Dict]

        def inner_func(conn):
            # type: (Connection) -> Dict[str, Dict]

            query = self._OrderItemsQuery

            order_list = []
            for orders in all_orders.values():
                order_list += orders
            order_list = ','.join(map(str, order_list))
            query = query.format(order_list)

            try:
                order_items = {}
                cursor = conn.select(query)
                for row in cursor:
                    order_id = str(row.get_entry(4))
                    product_code = str(row.get_entry(0))
                    quantity = float(row.get_entry(1))
                    if order_id in order_items:
                        order_items[order_id].append((product_code, quantity))
                    else:
                        order_items[order_id] = [(product_code, quantity)]

                for operator_id, orders_ids in all_orders.items():
                    if operator_id not in all_order_items.keys():
                        all_order_items[operator_id] = {}
                    for order_id in orders_ids:
                        for order_tuple in order_items[str(order_id)]:
                            product_code = order_tuple[0]
                            quantity = order_tuple[1]

                            if quantity == 0:
                                continue

                            if product_code not in all_order_items[operator_id].keys():
                                all_order_items[operator_id][product_code] = quantity
                            else:
                                all_order_items[operator_id][product_code] += quantity
                return all_order_items
            except Exception as e:
                sys_log_error("get order error: %s" % str(e))
                return {}

        return self.execute_with_connection(inner_func, db_name=str(self.pos_list[pos_index]))

    def get_paid_orders_by_business_period(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Dict[str, List]]
        return self._get_paid_orders(initial_date, end_date)

    def _get_paid_orders(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Dict[str, List]]
        def inner_func(conn):
            # type: (Connection) -> List[Dict[str, List]]

            try:
                query = self._PaidOrderQuery
                date_format = "%Y%m%d"

                query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))

                all_orders = {}  # type: Dict[str, List]
                valid_orders = [( int(x.get_entry(0)), x.get_entry(1) ) for x in conn.select(query)]

                for order_tuple in valid_orders:
                    order_id = order_tuple[0]
                    session_id = order_tuple[1]
                    session_id = session_id[session_id.find("user=") + 5:] if "user=" in session_id else ""
                    operator_id = session_id[0:session_id.find(",") if "," in session_id else None]

                    if operator_id not in all_orders:
                        all_orders[operator_id] = [order_id]
                    else:
                        all_orders[operator_id].append(order_id)

                return [all_orders]
            except Exception as _:
                return [{}]

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    _PaidOrderQuery = \
    """
        select distinct o.OrderId, o.SessionId
        from Orders o
        inner join OrderTender ot
        on o.OrderId = ot.OrderId
        where o.BusinessPeriod >= '{0}' and o.BusinessPeriod <= '{1}' and o.StateId = 5
    """

    _OrderItemsQuery = """\
    SELECT * FROM (
    SELECT PartCode, SUM(OrderedQty) as SOMA, ItemId, PriceListId1, OrderId
        FROM
            (SELECT 
			t.PartCode,
			CASE WHEN t.DefaultQty IS NULL
				THEN t.OrderedQty 
				ELSE
					CASE WHEN t.OrderedQty IS NULL
						THEN t.DefaultQty 
					ELSE
						CASE WHEN t.DefaultQty > t.OrderedQty
						THEN 0
						ELSE t.OrderedQty - t.DefaultQty
						END
					END
				END AS OrderedQty,
			t.ItemId,
			t.OrderId,
			s.PriceListId1
			FROM OrderItem t
			JOIN OrderItem u ON t.OrderId = u.OrderId AND t.LineNumber = u.LineNumber
			JOIN Orders s ON t.OrderId = s.OrderId
			JOIN ProductKernelParams pkp ON pkp.ProductCode = t.PartCode
			WHERE s.StateId = 5
			    AND u.Level = '0' 
			    AND t.OrderId IN ({0})
			    AND pkp.ProductType <> 1
			    AND t.PartCode NOT IN (
			        SELECT ProductCode FROM ProductTags
                    WHERE LOWER(Tag) = 'cfh=true'
                    UNION
                    SELECT ProductCode FROM ProductCustomParams
                    WHERE LOWER(CustomParamId) = 'cfh' AND LOWER(CustomParamValue) = 'true'
			    )
			    AND COALESCE(t.OrderedQty, t.DefaultQty) > 0) T
    GROUP BY PartCode, ItemId, OrderId) where SOMA is not null
    """



from actions.models import Navigation, NavigationProduct
from old_helper import BaseRepository
from persistence import Connection
from typing import List


class NavigationRepository(BaseRepository):
    def __init__(self, mb_context):
        super(NavigationRepository, self).__init__(mb_context)

    def get_navigation_list(self):
        # type: () -> List[Navigation]

        def inner_func(conn):
            # type: (Connection) -> List[Navigation]

            query = """
                        SELECT 
                           NV.NavId, 
                           Name, 
                           ParentNavId, 
                           SortOrder, 
                           ButtonText, 
                           COALESCE(ButtonColor, '#CCCCCC') AS ButtonColor,
                           NC.CustomParamValue AS RestrictedServicePeriod
                       FROM Navigation NV
                       LEFT JOIN NavigationCustomParams NC ON NC.NavId = NV.NavId 
                        AND CustomParamId = 'RESTRICTED_TO_SERVICE_PERIOD'
                       ORDER BY SortOrder
                   """

            res = list()
            for row in conn.select(query):
                nav = Navigation(nav_id=row.get_entry("NavId"),
                                 name=row.get_entry("Name"),
                                 parent_nav_id=row.get_entry("ParentNavId"),
                                 sort_order=row.get_entry("SortOrder"),
                                 button_text=row.get_entry("ButtonText"),
                                 background_color=row.get_entry("ButtonColor"),
                                 restricted_period=row.get_entry("RestrictedServicePeriod"))
                res.append(nav)

            return res

        return self.execute_with_connection(inner_func)

    def get_navigation_products(self):
        # type: () -> List[NavigationProduct]
        def inner_func(conn):
            # type: (Connection) -> List[NavigationProduct]

            query = """
                       SELECT PN.NavId, 
                              PN.ClassCode, 
                              PN.ProductCode, 
                              P.ProductName, 
                              COALESCE(N.ButtonColor, '#CCCCCC') AS ButtonColor
                       FROM ProductNavigation PN
                       INNER JOIN Product P ON PN.ProductCode = P.ProductCode
                       LEFT JOIN Navigation N ON PN.NavId = N.NavId
                       JOIN ProductKernelParams PKP ON P.ProductCode = PKP.ProductCode
                       ORDER BY PKP.ProductPriority
                    """

            res = list()
            for row in conn.select(query):
                nav = NavigationProduct(row.get_entry("NavId"),
                                        row.get_entry("ClassCode"),
                                        row.get_entry("ProductCode"),
                                        row.get_entry("ProductName"),
                                        row.get_entry("ButtonColor"))
                res.append(nav)

            return res

        return self.execute_with_connection(inner_func)

    def get_navigations_restricted_to_service_period(self, service_period_id):
        # type: () -> List[str]

        def inner_func(conn):
            # type: (Connection) -> List[str]

            query = """
                        SELECT NavId,
                               CustomParamValue
                          FROM NavigationCustomParams
                         WHERE CustomParamId = 'RESTRICTED_TO_SERVICE_PERIOD'
                   """

            res = []
            for row in conn.select(query):
                period_ids = row.get_entry("CustomParamValue").split(",")
                if str(service_period_id) in period_ids:
                    res.append(str(row.get_entry("NavId")))

            return res

        return self.execute_with_connection(inner_func)



from logging import getLogger
from typing import Optional

from nperipherals_core.model.device import ReaderDevice, DeviceServiceThread
from nperipherals_core.model.custom_exceptions import DeviceNotConnected
from nperipherals_core.model.repository import NotificationRepository

logger = getLogger(__name__)


class ReaderDeviceServiceThread(DeviceServiceThread):
    def __init__(self, device, notification_repository):
        # type: (ReaderDevice, NotificationRepository) -> None

        super(ReaderDeviceServiceThread, self).__init__()

        self.device = device
        self.notification_repository = notification_repository

        self.SLEEP_IN_SECONDS = 1

    def run(self):
        # type: () -> None

        reader_device_name = self.device.get_name()
        logger.info("Starting device [{}]".format(reader_device_name))

        while self.running:
            self._initialize(reader_device_name)
            self._connect(reader_device_name)
            self._read(reader_device_name)
            self._disconnect(reader_device_name)

    def _initialize(self, reader_device_name):
        # type: (unicode) -> None

        while self.running:
            try:
                logger.info("Initializing device [{}]".format(reader_device_name))
                self.device.initialize()
                break
            except Exception as _:
                logger.exception("Error initializing device [{}]".format(reader_device_name))
                self._wait()
                continue

    def _connect(self, reader_device_name):
        # type: (unicode) -> None

        while self.running:
            try:
                logger.info("Connecting device [{}]".format(reader_device_name))
                self.device.connect()
                logger.info("Connected to device [{}]".format(reader_device_name))
                break
            except Exception as _:
                logger.exception("Error connecting to device [{}]".format(reader_device_name))
                self._wait()
                continue

    def _read(self, reader_device_name):
        # type: (unicode) -> None

        while self.running:
            try:
                logger.debug("Reading from device [{}]".format(reader_device_name))
                data = self.device.read()

                logger.debug("Notifying data from device [{}]".format(reader_device_name))
                self.notification_repository.notify(reader_device_name, data)

                self._wait(0.1)
            except DeviceNotConnected as _:
                logger.warning("Error connecting to [{}]".format(reader_device_name))
                break
            except Exception as _:
                logger.warning("Error reading data from [{}]".format(reader_device_name))
                self._wait()
                continue

    def _disconnect(self, reader_device_name):
        # type: (unicode) -> None

        try:
            logger.info("Disconnecting device [{}]".format(reader_device_name))
            self.device.disconnect()
            logger.info("Disconnecting to device [{}]".format(reader_device_name))
        except Exception as _:
            logger.warning("Error Disconnecting to device [{}]".format(reader_device_name))

    def _wait(self, seconds=None):
        # type: (Optional[float]) -> None

        self.sleep(seconds or self.SLEEP_IN_SECONDS)



from production.box import RemoveReprocessedOrdersBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class RemoveReprocessedOrdersBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(RemoveReprocessedOrdersBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> RemoveReprocessedOrdersBox
        return RemoveReprocessedOrdersBox(config.name,
                                          config.sons,
                                          self.get_extra(config, "ReprocessedTimeLimit", None),
                                          self.get_logger(config))



from application.domain import TableService  # noqa
from application.domain.ui import ScreenChanger, ScreenIdEnum  # noqa
from application.interactor.service import TableReleaser  # noqa


class StartServiceInteractor(object):
    def __init__(self, screen_changer, table_service, table_releaser):
        # type: (ScreenChanger, TableService, TableReleaser) -> None
        self.screen_changer = screen_changer
        self.table_service = table_service
        self.table_releaser = table_releaser

    def execute(self, pos_id, table_id="", seats="", with_order=False, tab_id=""):
        # type: (str, str, int, bool) -> str

        self.table_releaser.release_all(pos_id)
        ret = self.table_service.start_service(pos_id, table_id, seats, tab_id)
        if with_order:
            self.table_service.create_order(pos_id, table_id)
            self.screen_changer.change(pos_id, ScreenIdEnum.create_table_order)
            return "REDIRECT"

        return ret



# -*- coding: utf-8 -*-
from msgbus import TK_POS_LISTUSERS, FM_PARAM, TK_SYS_NAK, MBEasyContext
from mw_helper import show_filtered_list_box_dialog, show_message_dialog
from sysactions import send_message, get_model, get_user_information, show_messagebox, StopAction
from xml.etree import cElementTree as eTree
from userhelper import UserHelper

from .. import USECS_PER_SEC, logger

mbcontext = None  # type: MBEasyContext


def list_opened_users(pos_id):
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, "%s" % pos_id, timeout=600 * USECS_PER_SEC)
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    model = get_model(pos_id)
    working_mode = model.find("WorkingMode").get("usrCtrlType")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("userControlType") == working_mode]

    if working_mode == 'QSR':
        working_users = [tag for tag in working_users if tag.get("posId") == pos_id]

    return working_users


def list_all_opened_users(pos_id):
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, "%s" % pos_id, timeout=600 * USECS_PER_SEC)

    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error listing logged operators" % int(pos_id))
        raise Exception("Error listing logged operators")

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users]

    return working_users


def list_ids_opened_users():
    with UserHelper(mbcontext) as user_helper:
        users_id = []
        users = user_helper.get_all_logged_users()
        for x in users:
            users_id.append(int(x))

        return users_id


def check_valid_user(pos_id, user_id):
    user_xml = get_user_information(user_id)
    if user_xml is None:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    user_list = eTree.XML(user_xml).findall("user")
    user_filter = filter(lambda x: x.get("UserId") == str(user_id), user_list)
    if user_filter in [None, [], ""]:
        show_message_dialog(pos_id, message="$INVALID_USER_OR_PASSWORD")
        return False

    return True


def get_opened_users(pos_id):
    opened_users = list_opened_users(pos_id)
    if not opened_users:
        show_message_dialog(pos_id, message="$THERE_ARE_NO_OPENED_USERS")
        return None, None

    index = 0
    if len(opened_users) > 1:
        opened_users.sort(key=lambda x: x.get("id"))

        parsed_users_options = []
        for user in opened_users:
            parsed_users_options.append("{} - {}".format(user.get('id'), user.get('name')))

        index = show_filtered_list_box_dialog(pos_id, "|".join(parsed_users_options), "$LOGOUT", "", "NOFILTER")
        if index is None:
            return None, None

    model = get_model(pos_id)
    operator = [op for op in model.findall("Operators/Operator") if op.get("id") == opened_users[index].get('id')]
    return opened_users[index], operator[0]


def is_user_already_logged(pos_id, user_id, block_by_working_mode=False):
    msg = send_message("POS%d" % int(pos_id), TK_POS_LISTUSERS, FM_PARAM, "%s" % pos_id, timeout=600 * USECS_PER_SEC)
    if msg.token == TK_SYS_NAK:
        logger.error("Logoff POS %d - Error getting logged users" % int(pos_id))
        return True

    xml = eTree.XML(msg.data)
    opened_users = [tag for tag in xml.findall("User") if not tag.get("closeTime")]
    working_users = [tag for tag in opened_users if tag.get("id") == user_id]

    if len(working_users) > 0:
        model = get_model(pos_id)

        if block_by_working_mode:
            user_info = [x for x in opened_users if x.get("id") == user_id]
            if user_info:
                pos_working_mode = model.find("WorkingMode").get("usrCtrlType")
                operator_working_mode = "QSR" if "pos" in user_info[0].get("sessionId") else "TS"
                if pos_working_mode != operator_working_mode:
                    show_messagebox(pos_id, "$USER_LOGGED_IN_ANOTHER_POS_TYPE|{}".format(user_info[0].get("posId")))
                    raise StopAction()

        operator = [op for op in model.findall("Operators/Operator") if op.get("id") == user_id]

        if len(operator) > 0 and operator[0].get('state') != 'LOGGEDIN':
            return False

        user_info = [x for x in opened_users if x.get("id") == user_id]
        show_message_dialog(pos_id, message="$OPERATOR_ALREADY_LOGGED|{}".format(user_info[0].get("posId")))
        return True

    return False



from _Generator import Generator  # noqa
from _Formatter import Formatter  # noqa
from commons.util import Translator


class Report(object):
    def __init__(self, generator, formatter):
        # type:(Generator, Formatter) -> None
        self.generator = generator
        self.formatter = formatter
        self.translator = Translator()

    def generate_and_format_report(self):
        dto = self.generator.generate_data()
        return self.formatter.format_report(dto, self.translator)



from typing import Dict, List, Union

from nperipherals_core.model.device import ReaderDevice, WriterDevice
from nperipherals_core.model.custom_exceptions import UnknownDevice


class DeviceFactory(object):
    def __init__(self, device_types):
        # type: (List[type]) -> None

        self.handled_devices = dict()  # type: Dict[str, type]

        for device_type in device_types:
            self._register(device_type)

    def create_device(self, device_configurations):
        # type: (Dict) -> Union[ReaderDevice, WriterDevice]

        if not device_configurations or "Type" not in device_configurations:
            raise UnknownDevice()

        device_type_str = device_configurations["Type"]  # type: str

        if device_type_str in self.handled_devices:
            return self.handled_devices[device_type_str](device_configurations)

        raise UnknownDevice(device_type_str)

    def _register(self, device_type):
        # type: (type) -> None

        self.handled_devices[device_type.__name__] = device_type



from ._OperatorNameRetriever import OperatorNameRetriever
from ._Session import Session
from ._SessionRetriever import SessionRetriever
from ._Store import Store
from ._StoreRetriever import StoreRetriever
from ._Order import Order
from ._OrderTender import OrderTender
from ._OrderRepository import OrderRepository
from ._Transfer import Transfer
from ._TransferRepository import TransferRepository
from ._SessionRepository import SessionRepository
from ._Translator import Translator
from ._L10n import L10n
from basereport._I18n import I18n
from ._Clock import Clock
from ._TableService import TableService
from ._TableOrderRetriever import TableOrderRetriever
from ._TableOrderSaleLine import TableOrderSaleLine
from ._TableOrder import TableOrder
from ._Table import Table
from ._OrderStateEnum import OrderStateEnum
from ._TipRepository import TipRepository
from ._OperatorTip import OperatorTip
from ._TransferTypeEnum import TransferType

__all__ = [
    "OperatorNameRetriever",
    "Session",
    "StoreRetriever",
    "Store",
    "StoreRetriever",
    "Order",
    "OrderTender",
    "OrderRepository",
    "Transfer",
    "TransferRepository",
    "SessionRepository",
    "Translator",
    "L10n",
    "I18n",
    "Clock",
    "TableService",
    "TableOrderRetriever",
    "TableOrderSaleLine",
    "TableOrder",
    "OrderStateEnum",
    "Table",
    "TipRepository",
    "OperatorTip",
    "TransferType"
]



from typing import Union, Optional

from messagebus import MessageBus

from externals.repository import MessageBusRepository
from nperipherals_core.service import ReaderDeviceServiceThread
from nperipherals_core.model.device import Device, ReaderDevice, WriterDevice


class DeviceServiceThreadFactory(object):
    def __init__(self, message_bus):
        # type: (MessageBus) -> None

        self.message_bus = message_bus

    def create_device_service_thread(self, device):
        # type: (Union[Device, ReaderDevice, WriterDevice]) -> Optional[ReaderDeviceServiceThread]

        device_type = type(device)  # type: str

        if issubclass(device_type, ReaderDevice):
            return ReaderDeviceServiceThread(device, notification_repository=MessageBusRepository(self.message_bus))



# encoding: utf-8
import sys

from msgbus import MBException, MBEasyContext, SE_SUCCESS, TK_CMP_TERM_NOW, TK_CMP_START, TK_EVT_EVENT, TK_SYS_NAK, TK_SYS_ACK, MBMessage
from syserrors import SE_MB_ERR
from systools import sys_log_exception, sys_log_info
from threading import Thread, Lock, Condition
from typing import List, Dict, Optional


class MessageHandler:
    def __init__(self, mbcontext, service_name, service_type, required_services, event_handler=None):
        # type: (MBEasyContext, str, str, str, Optional[EventHandler]) -> None
        self.mbcontext = mbcontext
        self.service_name = service_name
        self.service_type = service_type
        self.required_services = required_services
        self.event_handler = event_handler
        self.reentrant_subjects = {}
        self.sync_subjects = {}
        self.non_reentrant_subjects = {}
        self.queue_subjects = {}
        self.thread_handling_subject = {}
        self.subject_event_queue = {}  # type: Dict[unicode, List[tuple]]
        self.subject_queue_thread = {}  # type: Dict[unicode, Thread]
        self.subject_queue_condition = {}  # type: Dict[unicode, Condition]

        self.tokens = {}
        self.finished = False

        try:
            # Se temos um servico, registramos o mesmo
            if self.service_name and self.service_type != "":
                status = self.mbcontext.MB_EasyWaitStart("%s:%s" % (self.service_name, self.service_type), self.required_services)
                if status != SE_SUCCESS:
                    sys_log_info("Terminating component after 'MB_EasyWaitStart'")
                    sys.exit(status)
        except MBException:
            sys_log_exception("Message-bus error while initializing component. Terminating")
            sys.exit(SE_MB_ERR)

        if self.event_handler:
            self.subscribe_tokens(self.event_handler.get_handled_tokens())

    def set_event_handler(self, event_handler):
        self.event_handler = event_handler
        self.subscribe_tokens(self.event_handler.get_handled_tokens())

    def subscribe_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MessageHandler
        for subject in subjects:
            self.reentrant_subjects[subject] = subject
            self.mbcontext.MB_EasyEvtSubscribe(subject)

        return self

    def subscribe_non_reentrant_events(self, subjects):
        # type: (List[unicode]) -> MessageHandler
        for subject in subjects:
            self.non_reentrant_subjects[subject] = subject
            self.mbcontext.MB_EasyEvtSubscribe(subject)

        return self

    def subscribe_queue_events(self, subjects):
        for subject in subjects:
            self.queue_subjects[subject] = subject

            self.subject_event_queue[subject] = []
            self.subject_queue_condition[subject] = Condition()
            subject_queue_thread = Thread(target=self.event_queue_method, name=subject + " - Event Queue Thread", args=(subject,))
            subject_queue_thread.daemon = True
            subject_queue_thread.start()
            self.subject_queue_thread[subject] = subject_queue_thread

            self.mbcontext.MB_EasyEvtSubscribe(subject)

        return self

    def subscribe_sync_events(self, subjects):
        for subject in subjects:
            self.sync_subjects[subject] = subject
            self.mbcontext.MB_EasyEvtSubscribe(subject)

        return self

    def subscribe_events(self, subjects):
        # type: (List[unicode]) -> MessageHandler
        return self.subscribe_non_reentrant_events(subjects)

    def subscribe_tokens(self, tokens):
        # type: (list) -> MessageHandler
        for token in tokens:
            self.tokens[token] = token

        return self

    def handle_events(self):
        while not self.finished:
            msg = self.mbcontext.MB_EasyGetMessage()

            # Check for component terminate request
            if (not msg) or (msg.token == TK_CMP_TERM_NOW):
                # Component terminated, warn the handler that we are terminated
                self.event_handler.terminate_event()
                # reply the message
                self.mbcontext.MB_EasyReplyMessage(msg)
                # Finish the loop and terminate the component
                self.finished = True
                break

            if msg.token == TK_CMP_START:
                msg.token = TK_SYS_ACK
                self.mbcontext.MB_EasyReplyMessage(msg)

            elif msg.token == TK_EVT_EVENT:
                params = msg.data.split('\x00')
                data = params[0] if len(params) > 0 else None
                subject = params[1] if len(params) > 1 else None
                evt_type = params[2] if len(params) > 2 else None

                if subject in self.non_reentrant_subjects and subject in self.thread_handling_subject and self.thread_handling_subject[subject].is_alive():
                    # Se o evento é não reentrante e já temos uma thread tratando outra instância do mesmo evento, respondemos e não criamos outra thread
                    self.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                if subject in self.queue_subjects:
                    # Se o evento é do tipo fila, enfileiramos a tupla e deixamos a thread responsável tratá-lo
                    event = (subject, evt_type, data, msg)
                    with self.subject_queue_condition[subject]:
                        self.subject_event_queue[subject].append(event)
                        self.subject_queue_condition[subject].notify()

                    self.mbcontext.MB_EasyReplyMessage(msg)
                    continue

                handler_thread = Thread(target=self.event_handler.handle_event, args=(subject, evt_type, data, msg))
                handler_thread.daemon = True

                if subject in self.non_reentrant_subjects:
                    self.thread_handling_subject[subject] = handler_thread

                handler_thread.start()

                # Se o evento for syncrono, deixamos a thread responsável por dar o reply na mensagem
                if subject not in self.sync_subjects:
                    self.mbcontext.MB_EasyReplyMessage(msg)

            # Check for a event request
            elif msg.token in self.tokens:
                # handles the event ...
                handler_thread = Thread(target=self.event_handler.handle_message, args=(msg,))
                handler_thread.daemon = True
                handler_thread.start()
            else:
                # Ignore any unknown message (just reply it with a TK_SYS_NAK)
                msg.token = TK_SYS_NAK
                self.mbcontext.MB_EasyReplyMessage(msg)

    def event_queue_method(self, subject):
        my_condition = self.subject_queue_condition[subject]
        my_queue = self.subject_event_queue[subject]

        while not self.finished:
            while len(my_queue) > 0:
                with my_condition:
                    event = my_queue.pop(0)

                subject = event[0]
                evt_type = event[1]
                data = event[2]
                msg = event[3]
                try:
                    self.event_handler.handle_event(subject, evt_type, data, msg)
                except:
                    pass

            with my_condition:
                my_condition.wait()


class EventHandler(object):
    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        self.mbcontext = mbcontext

    def get_handled_tokens(self):
        raise NotImplementedError()

    def handle_message(self, msg):
        raise NotImplementedError()

    def handle_event(self, subject, evt_type, data, msg):
        # type: (unicode, unicode, str, MBMessage) -> None
        raise NotImplementedError()

    def terminate_event(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
import json
from datetime import datetime

from actions.models import GLAccount, TransferType, CloseOperator
from old_helper import BaseRepository


class AccountRepository(BaseRepository):
    def __init__(self, mb_context):
        super(AccountRepository, self).__init__(mb_context)

    def insert_close_operator_info(self, operator):
        # type: (CloseOperator) -> float

        def inner_func(conn):
            total_bordereau_difference = 0
            initial_fund = self.get_initial_amount(operator.pos_id, operator.session)
            total_cash_in = self.get_cash_in_total_amount(operator.pos_id, operator.session)
            total_cash_out = self.get_cash_out_total_amount(operator.pos_id, operator.session)
            final_cash = initial_fund + total_cash_in - total_cash_out
            for value in operator.values:
                amount = float(value["amount"])
                tender_id = int(value["tenderId"])
                has_payment = str(tender_id) in operator.payments
                sold_amount = 0.0 if not has_payment else operator.payments[str(tender_id)]
                envelope = _get_envelope(tender_id, operator)
                total_bordereau_difference += _get_total_difference(tender_id, amount, sold_amount, final_cash)

                gl_account = GLAccount(manager=operator.manager_id, envelope=envelope)
                gl_account_json = json.dumps(gl_account, default=lambda o: o.__dict__, sort_keys=True)

                conn.query("""INSERT INTO Transfer VALUES ('{}', {}, '{}', {}, '{}', {}, {}, '{}', '{}')"""
                           .format(operator.period, operator.pos_id, operator.session, TransferType.DECLARED_AMOUNT,
                                   "DECLARED_AMOUNT", amount, tender_id,
                                   datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"), gl_account_json))

                conn.query("""INSERT INTO Transfer VALUES ('{}', {}, '{}', {}, '{}', {}, {}, '{}', '{}')"""
                           .format(operator.period, operator.pos_id, operator.session, TransferType.TRANSFER_SKIM,
                                   "TRANSFER_SKIM", sold_amount, tender_id,
                                   datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"), ""))

            return total_bordereau_difference

        return self.execute_with_transaction(inner_func, db_name=str(operator.pos_id))

    def update_justification(self, pos_id, session, justification):
        def inner_func(conn):
            conn.query("""UPDATE Transfer
                          SET GLAccount = REPLACE(GLAccount, '"justification": ""', '"justification": "{}"')
                          WHERE SessionId = '{}' AND GLAccount IS NOT NULL AND GLAccount != '' AND Type IN (5,6)"""
                       .format(justification, session))

        self.execute_with_transaction(inner_func, db_name=str(pos_id))

    def get_closing_operator_sessions(self, pos_id, user, start_date, end_date):
        def inner_func(conn):
            sessions = []
            query = """SELECT SessionId
                       FROM Transfer
                       WHERE SessionId LIKE '%user={},%' and Period >= '{}' and Period <= '{}'
                       GROUP BY SessionId
                       ORDER BY Timestamp""".format(user, start_date, end_date)

            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    sessions.append(row.get_entry("SessionId"))

            return sessions

        return self.execute_with_connection(inner_func, db_name=str(pos_id))

    def get_initial_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.INITIAL_AMOUNT)

    def get_cash_in_total_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.TRANSFER_CASH_IN)

    def get_cash_out_total_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.TRANSFER_CASH_OUT)

    def get_transfer_total_amount(self, pos_id, session_id, transfer_type):
        def inner_func(conn):
            query = """
                        SELECT amount FROM transfer
                        WHERE Type = '{}' AND SessionId = '{}'
                    """.format(transfer_type, session_id)
            cursor = conn.select(query)
            if cursor.rows() > 0:
                total_cash_in = 0
                for row in cursor:
                    total_cash_in += float(row.get_entry(0) or 0.0)

                return total_cash_in

            return 0

        return self.execute_with_connection(inner_func, db_name=str(pos_id))


def _get_envelope(tender_id, operator):
    if tender_id == 0:
        return operator.envelope
    return ""


def _get_total_difference(tender_id, amount, sold_amount, final_cash):
    if tender_id == 0:
        return (amount - sold_amount) - final_cash

    return amount - sold_amount



from ._NfeCardBrand import NfeCardBrand
from ._NfeTender import NfeTender
from ._PosTenderBrandToNfeCardBrand import PosTenderBrandToNfeCardBrand
from ._PosTenderToNfeTender import PosTenderToNfeTender

from softwaredeveloper import SoftwareDeveloper

__all__ = [
    "SoftwareDeveloper",
    "PosTenderToNfeTender",
    "PosTenderBrandToNfeCardBrand",
    "NfeTender",
    "NfeCardBrand"
]



# -*- coding: utf-8 -*-
import base64
import gzip
import json
import threading
import requests
from cStringIO import StringIO
from xml.etree import cElementTree as eTree

from application.configurator import Configurator
from application.eventsender import EventSender
from application.interactor import OrderPictureInteractor, OperatorLogoutInteractor, ProductionOrderInteractor, \
    TablePictureInteractor
from application.model import EventData, AuditEvent, Command
from ._AuditEventConsumer import AuditEventConsumer
from syserrors import SE_SUCCESS, SE_BADXML, SE_BADRESPONSE
from typing import Optional


class AuditEventConsumerEventHandler(AuditEventConsumer):

    def __init__(self, configurations):
        # type: (Configurator) -> None

        self.configs = configurations
        self.repository = None
        self.event_sender = EventSender(self.configs)

        self.operator_logout_interactor = None
        self.order_picture_interactor = None
        self.table_picture_interactor = None
        self.production_order_interactor = None

        super(self.__class__, self).__init__(self.configs, self.process)

        with self.lock:
            self._create_commands_table()

        t = threading.Thread(target=self._status_sender)
        t.daemon = True
        t.start()

    def set_repository(self, repository):
        self.repository = repository

    def build_interactors(self):
        # type: () -> None

        self.operator_logout_interactor = OperatorLogoutInteractor(self.configs, self.repository)
        self.order_picture_interactor = OrderPictureInteractor(self.configs, self.repository)
        self.table_picture_interactor = TablePictureInteractor(self.configs)
        self.production_order_interactor = ProductionOrderInteractor(self.configs)

    def process(self, evt_list):
        for evt in evt_list:
            try:
                data, evt_type = evt[0], evt[1]
                self.logger.info("New event received: [{}]".format(evt_type))
                event_data = self._get_event_data(evt_type, data)
            except Exception as _:
                message = "Auditlog event parser error. Exiting to force resynchronization. Event: {}".format(evt)
                self.logger.exception(message)
                return SE_BADXML

            try:
                if evt_type in [AuditEvent.PAID, AuditEvent.VOID_ORDER, AuditEvent.ORDER_REPROCESS]:
                    payload = self.order_picture_interactor.build(event_data)
                    result = self.event_sender.send_request(event_data, payload)
                    if result != SE_SUCCESS:
                        return result

                elif evt_type in [AuditEvent.TABLE_CLOSED, ]:
                    payload = self.table_picture_interactor.build(event_data)
                    result = self.event_sender.send_request(event_data, payload, True)
                    if result != SE_SUCCESS:
                        return result

                elif evt_type == AuditEvent.OPERATOR_LOGOUT:
                    payload = self.operator_logout_interactor.build(event_data)
                    if not payload:
                        continue
                    result = self.event_sender.send_request(event_data, payload)
                    if result != SE_SUCCESS:
                        return result

                elif evt_type == AuditEvent.PRODUCTION_ORDER_TIMES:
                    payload = self.production_order_interactor.build(event_data)
                    result = self.event_sender.send_request(event_data, payload)
                    if result != SE_SUCCESS:
                        return result

            except Exception as _:
                self.logger.exception("[{}] Unable to process event".format(evt_type))
                return SE_BADRESPONSE

        return SE_SUCCESS

    def _status_sender(self):
        ctrl_conn = self._create_db_conn()

        while not self.terminating:
            wait_time = 15
            try:
                command = None
                if self.repository is not None:
                    order_picture = self.repository.get_last_order_picture()
                    last_command_id = self._get_last_command_id(ctrl_conn)
                    command = self._send_status(order_picture, last_command_id)
                    wait_time = 600

                if command and command.type == "ReprocessSales":
                    self.logger.info("Received ReprocessSales Command")
                    self._update_command_id(command.id, ctrl_conn)
                    self.force_new_period(command.data, ctrl_conn)

            except WindowsError as _:
                self.renew_service()
                self.logger.exception("On WindowsError")
            except Exception as _:
                wait_time = 60
                self.logger.exception("ERROR on PUMP Status Thread")

            finally:
                with self.condition:
                    self.condition.wait(wait_time)

    def _create_commands_table(self, ctrl_conn=None):
        if not ctrl_conn:
            ctrl_conn = self.ctrl_conn

        ctrl_conn.execute("""
                CREATE TABLE IF NOT EXISTS CommandsProcessed (
                    CommandId   INTEGER NOT NULL,
                    PRIMARY KEY (CommandId)
                );
            """)

    def _update_command_id(self, command_id, ctrl_conn=None):
        if not ctrl_conn:
            ctrl_conn = self.ctrl_conn

        query = "INSERT OR IGNORE INTO CommandsProcessed(CommandId) VALUES(:command_id);"
        ctrl_conn.execute(query, {"command_id": command_id})

    def _get_last_command_id(self, ctrl_conn=None):
        if not ctrl_conn:
            ctrl_conn = self.ctrl_conn

        c = ctrl_conn.cursor()
        r = c.execute("""SELECT MAX(CommandId) FROM CommandsProcessed;""").fetchone()
        return r[0]

    def _send_status(self, order_picture, last_command_id):
        # type: (str, str) -> Optional[Command]

        xml_picture = eTree.fromstring(order_picture)
        order = xml_picture.find(".//Order")
        if order is None or order.get("orderId") is None:
            raise Exception()

        message = {
            "lastOrderId": order.get("orderId"),
            "lastOrderDatetime": order.get("createdAtGMT") + "Z",
            "lastCommandIdProcessed": last_command_id
        }

        payload = json.dumps(message)
        try:
            payload = payload.encode("utf-8")
            self.logger.info("Sending Pump Status to server. Payload:\n{}".format(payload))

            headers = {
                "Accept": "application/json",
                "Content-type": "application/json",
                "x-api-key": self.configs.api_key
            }

            post_url = self.configs.server_url + self.configs.pump_status_url + "/{}".format(self.configs.store_id)
            response = requests.post(post_url, headers=headers, data=payload, timeout=30, verify=False)  # NOSONAR

            if response.status_code != 200:
                log_message = "Error sending Pump Status server. Status: {} - {}"
                self.logger.info(log_message.format(response.status_code, response.content))
                return None

            self.logger.info("Success sending Pump Status to server")

            if response.content == "":
                return None
            return Command(response.json())

        except Exception as _:
            message = "Communication error sendingPump Status to server"
            self.logger.exception(message)
            return None

    @staticmethod
    def _get_event_data(evt_type, xml):
        # type: (str, str) -> EventData

        event = eTree.XML(xml)
        audit_log_evt = event.find("AuditLog")
        evt_timestamp = audit_log_evt.get("evtTime") or ""
        evt_data = audit_log_evt.text.encode("UTF-8")
        data = eTree.fromstring(gzip.GzipFile(fileobj=StringIO(base64.b64decode(evt_data))).read())
        session_id = data.get("sessionid")

        return EventData(data=data, event_type=evt_type, timestamp=evt_timestamp, session_id=session_id)



from ._Benefit import Benefit
from ._Action import Action
from ._Conditions import Conditions
from ._RequiredItemOnSale import RequiredItemOnSale
from ._RequiredGroupOnSale import RequiredGroupOnSale
from ._ActionTypes import ActionTypes
from ._OrderBenefitInfo import OrderBenefitInfo
from ._ActionResult import ActionResult
from ._BenefitList import BenefitList
from ._Exclusions import Exclusions



# -*- coding: utf-8 -*-

from datetime import datetime

from typing import Optional


class MessageFrom(object):
    SAC = "Sac"
    STORE = "Store"


class Message(object):
    def __init__(self):
        self.id = None  # type: Optional[int]
        self.message_from = None  # type: Optional[unicode]
        self.created_time = None  # type: Optional[datetime]
        self.received_time = None  # type: Optional[datetime]
        self.server_id = None  # type: Optional[unicode]
        self.text = None  # type: Optional[unicode]



import persistence
from systools import sys_log_exception

from .. import mb_context


available_measure_unit = 'kg'
_products_with_measure_unit_cache = []


def get_products_with_measure_unit_cache(pos_id='1'):
    global _products_with_measure_unit_cache

    if not _products_with_measure_unit_cache:
        conn = None
        try:
            conn = persistence.Driver().open(mb_context, pos_id)
            query = """
                        SELECT ProductCode 
                        FROM ProductKernelParams 
                        WHERE MeasureUnit = '{}' COLLATE NOCASE
                    """.format(available_measure_unit)
            cursor = conn.select(query)
            for row in cursor:
                product_code = row.get_entry(0)
                if product_code not in _products_with_measure_unit_cache:
                    _products_with_measure_unit_cache.append(product_code)

        except (persistence.AprException, persistence.DbdException) as _:
            sys_log_exception("Error getting products with measure unit")
        finally:
            if conn:
                conn.close()

    return _products_with_measure_unit_cache


def is_a_product_with_measure_unit(product_code):
    if str(product_code) in _products_with_measure_unit_cache:
        return True
    return False



from helper import ExtendedEnum
from messagebus import TokenCreator, TokenPriority

MESSAGE_GROUP = "113"


class ListenedTokens(ExtendedEnum):

    TK_NCOUPON_APPLY_COUPON = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_NCOUPON_UNLOCK_COUPON = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")
    TK_NCOUPON_BURN_COUPON = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "3")
    TK_NCOUPON_GET_COUPON_INFO = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "4")



from abc import ABCMeta, abstractmethod
from ._Table import Table  # noqa


class TableOrderRetriever(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_table_order(self, pos_id, table_id):
        # type: (str, str) -> Table

        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    flask.ctx
    ~~~~~~~~~

    Implements the objects required to keep the context.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
import sys
from functools import update_wrapper

from werkzeug.exceptions import HTTPException

from ._compat import BROKEN_PYPY_CTXMGR_EXIT
from ._compat import reraise
from .globals import _app_ctx_stack
from .globals import _request_ctx_stack
from .signals import appcontext_popped
from .signals import appcontext_pushed


# a singleton sentinel value for parameter defaults
_sentinel = object()


class _AppCtxGlobals(object):
    """A plain object. Used as a namespace for storing data during an
    application context.

    Creating an app context automatically creates this object, which is
    made available as the :data:`g` proxy.

    .. describe:: 'key' in g

        Check whether an attribute is present.

        .. versionadded:: 0.10

    .. describe:: iter(g)

        Return an iterator over the attribute names.

        .. versionadded:: 0.10
    """

    def get(self, name, default=None):
        """Get an attribute by name, or a default value. Like
        :meth:`dict.get`.

        :param name: Name of attribute to get.
        :param default: Value to return if the attribute is not present.

        .. versionadded:: 0.10
        """
        return self.__dict__.get(name, default)

    def pop(self, name, default=_sentinel):
        """Get and remove an attribute by name. Like :meth:`dict.pop`.

        :param name: Name of attribute to pop.
        :param default: Value to return if the attribute is not present,
            instead of raise a ``KeyError``.

        .. versionadded:: 0.11
        """
        if default is _sentinel:
            return self.__dict__.pop(name)
        else:
            return self.__dict__.pop(name, default)

    def setdefault(self, name, default=None):
        """Get the value of an attribute if it is present, otherwise
        set and return a default value. Like :meth:`dict.setdefault`.

        :param name: Name of attribute to get.
        :param: default: Value to set and return if the attribute is not
            present.

        .. versionadded:: 0.11
        """
        return self.__dict__.setdefault(name, default)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __repr__(self):
        top = _app_ctx_stack.top
        if top is not None:
            return "<flask.g of %r>" % top.app.name
        return object.__repr__(self)


def after_this_request(f):
    """Executes a function after this request.  This is useful to modify
    response objects.  The function is passed the response object and has
    to return the same or a new one.

    Example::

        @app.route('/')
        def index():
            @after_this_request
            def add_header(response):
                response.headers['X-Foo'] = 'Parachute'
                return response
            return 'Hello World!'

    This is more useful if a function other than the view function wants to
    modify a response.  For instance think of a decorator that wants to add
    some headers without converting the return value into a response object.

    .. versionadded:: 0.9
    """
    _request_ctx_stack.top._after_request_functions.append(f)
    return f


def copy_current_request_context(f):
    """A helper function that decorates a function to retain the current
    request context.  This is useful when working with greenlets.  The moment
    the function is decorated a copy of the request context is created and
    then pushed when the function is called.  The current session is also
    included in the copied request context.

    Example::

        import gevent
        from flask import copy_current_request_context

        @app.route('/')
        def index():
            @copy_current_request_context
            def do_some_work():
                # do some work here, it can access flask.request or
                # flask.session like you would otherwise in the view function.
                ...
            gevent.spawn(do_some_work)
            return 'Regular response'

    .. versionadded:: 0.10
    """
    top = _request_ctx_stack.top
    if top is None:
        raise RuntimeError(
            "This decorator can only be used at local scopes "
            "when a request context is on the stack.  For instance within "
            "view functions."
        )
    reqctx = top.copy()

    def wrapper(*args, **kwargs):
        with reqctx:
            return f(*args, **kwargs)

    return update_wrapper(wrapper, f)


def has_request_context():
    """If you have code that wants to test if a request context is there or
    not this function can be used.  For instance, you may want to take advantage
    of request information if the request object is available, but fail
    silently if it is unavailable.

    ::

        class User(db.Model):

            def __init__(self, username, remote_addr=None):
                self.username = username
                if remote_addr is None and has_request_context():
                    remote_addr = request.remote_addr
                self.remote_addr = remote_addr

    Alternatively you can also just test any of the context bound objects
    (such as :class:`request` or :class:`g`) for truthness::

        class User(db.Model):

            def __init__(self, username, remote_addr=None):
                self.username = username
                if remote_addr is None and request:
                    remote_addr = request.remote_addr
                self.remote_addr = remote_addr

    .. versionadded:: 0.7
    """
    return _request_ctx_stack.top is not None


def has_app_context():
    """Works like :func:`has_request_context` but for the application
    context.  You can also just do a boolean check on the
    :data:`current_app` object instead.

    .. versionadded:: 0.9
    """
    return _app_ctx_stack.top is not None


class AppContext(object):
    """The application context binds an application object implicitly
    to the current thread or greenlet, similar to how the
    :class:`RequestContext` binds request information.  The application
    context is also implicitly created if a request context is created
    but the application is not on top of the individual application
    context.
    """

    def __init__(self, app):
        self.app = app
        self.url_adapter = app.create_url_adapter(None)
        self.g = app.app_ctx_globals_class()

        # Like request context, app contexts can be pushed multiple times
        # but there a basic "refcount" is enough to track them.
        self._refcnt = 0

    def push(self):
        """Binds the app context to the current context."""
        self._refcnt += 1
        if hasattr(sys, "exc_clear"):
            sys.exc_clear()
        _app_ctx_stack.push(self)
        appcontext_pushed.send(self.app)

    def pop(self, exc=_sentinel):
        """Pops the app context."""
        try:
            self._refcnt -= 1
            if self._refcnt <= 0:
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                self.app.do_teardown_appcontext(exc)
        finally:
            rv = _app_ctx_stack.pop()
        assert rv is self, "Popped wrong app context.  (%r instead of %r)" % (rv, self)
        appcontext_popped.send(self.app)

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pop(exc_value)

        if BROKEN_PYPY_CTXMGR_EXIT and exc_type is not None:
            reraise(exc_type, exc_value, tb)


class RequestContext(object):
    """The request context contains all request relevant information.  It is
    created at the beginning of the request and pushed to the
    `_request_ctx_stack` and removed at the end of it.  It will create the
    URL adapter and request object for the WSGI environment provided.

    Do not attempt to use this class directly, instead use
    :meth:`~flask.Flask.test_request_context` and
    :meth:`~flask.Flask.request_context` to create this object.

    When the request context is popped, it will evaluate all the
    functions registered on the application for teardown execution
    (:meth:`~flask.Flask.teardown_request`).

    The request context is automatically popped at the end of the request
    for you.  In debug mode the request context is kept around if
    exceptions happen so that interactive debuggers have a chance to
    introspect the data.  With 0.4 this can also be forced for requests
    that did not fail and outside of ``DEBUG`` mode.  By setting
    ``'flask._preserve_context'`` to ``True`` on the WSGI environment the
    context will not pop itself at the end of the request.  This is used by
    the :meth:`~flask.Flask.test_client` for example to implement the
    deferred cleanup functionality.

    You might find this helpful for unittests where you need the
    information from the context local around for a little longer.  Make
    sure to properly :meth:`~werkzeug.LocalStack.pop` the stack yourself in
    that situation, otherwise your unittests will leak memory.
    """

    def __init__(self, app, environ, request=None, session=None):
        self.app = app
        if request is None:
            request = app.request_class(environ)
        self.request = request
        self.url_adapter = None
        try:
            self.url_adapter = app.create_url_adapter(self.request)
        except HTTPException as e:
            self.request.routing_exception = e
        self.flashes = None
        self.session = session

        # Request contexts can be pushed multiple times and interleaved with
        # other request contexts.  Now only if the last level is popped we
        # get rid of them.  Additionally if an application context is missing
        # one is created implicitly so for each level we add this information
        self._implicit_app_ctx_stack = []

        # indicator if the context was preserved.  Next time another context
        # is pushed the preserved context is popped.
        self.preserved = False

        # remembers the exception for pop if there is one in case the context
        # preservation kicks in.
        self._preserved_exc = None

        # Functions that should be executed after the request on the response
        # object.  These will be called before the regular "after_request"
        # functions.
        self._after_request_functions = []

    @property
    def g(self):
        return _app_ctx_stack.top.g

    @g.setter
    def g(self, value):
        _app_ctx_stack.top.g = value

    def copy(self):
        """Creates a copy of this request context with the same request object.
        This can be used to move a request context to a different greenlet.
        Because the actual request object is the same this cannot be used to
        move a request context to a different thread unless access to the
        request object is locked.

        .. versionadded:: 0.10

        .. versionchanged:: 1.1
           The current session object is used instead of reloading the original
           data. This prevents `flask.session` pointing to an out-of-date object.
        """
        return self.__class__(
            self.app,
            environ=self.request.environ,
            request=self.request,
            session=self.session,
        )

    def match_request(self):
        """Can be overridden by a subclass to hook into the matching
        of the request.
        """
        try:
            result = self.url_adapter.match(return_rule=True)
            self.request.url_rule, self.request.view_args = result
        except HTTPException as e:
            self.request.routing_exception = e

    def push(self):
        """Binds the request context to the current context."""
        # If an exception occurs in debug mode or if context preservation is
        # activated under exception situations exactly one context stays
        # on the stack.  The rationale is that you want to access that
        # information under debug situations.  However if someone forgets to
        # pop that context again we want to make sure that on the next push
        # it's invalidated, otherwise we run at risk that something leaks
        # memory.  This is usually only a problem in test suite since this
        # functionality is not active in production environments.
        top = _request_ctx_stack.top
        if top is not None and top.preserved:
            top.pop(top._preserved_exc)

        # Before we push the request context we have to ensure that there
        # is an application context.
        app_ctx = _app_ctx_stack.top
        if app_ctx is None or app_ctx.app != self.app:
            app_ctx = self.app.app_context()
            app_ctx.push()
            self._implicit_app_ctx_stack.append(app_ctx)
        else:
            self._implicit_app_ctx_stack.append(None)

        if hasattr(sys, "exc_clear"):
            sys.exc_clear()

        _request_ctx_stack.push(self)

        # Open the session at the moment that the request context is available.
        # This allows a custom open_session method to use the request context.
        # Only open a new session if this is the first time the request was
        # pushed, otherwise stream_with_context loses the session.
        if self.session is None:
            session_interface = self.app.session_interface
            self.session = session_interface.open_session(self.app, self.request)

            if self.session is None:
                self.session = session_interface.make_null_session(self.app)

        if self.url_adapter is not None:
            self.match_request()

    def pop(self, exc=_sentinel):
        """Pops the request context and unbinds it by doing that.  This will
        also trigger the execution of functions registered by the
        :meth:`~flask.Flask.teardown_request` decorator.

        .. versionchanged:: 0.9
           Added the `exc` argument.
        """
        app_ctx = self._implicit_app_ctx_stack.pop()

        try:
            clear_request = False
            if not self._implicit_app_ctx_stack:
                self.preserved = False
                self._preserved_exc = None
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                self.app.do_teardown_request(exc)

                # If this interpreter supports clearing the exception information
                # we do that now.  This will only go into effect on Python 2.x,
                # on 3.x it disappears automatically at the end of the exception
                # stack.
                if hasattr(sys, "exc_clear"):
                    sys.exc_clear()

                request_close = getattr(self.request, "close", None)
                if request_close is not None:
                    request_close()
                clear_request = True
        finally:
            rv = _request_ctx_stack.pop()

            # get rid of circular dependencies at the end of the request
            # so that we don't require the GC to be active.
            if clear_request:
                rv.request.environ["werkzeug.request"] = None

            # Get rid of the app as well if necessary.
            if app_ctx is not None:
                app_ctx.pop(exc)

            assert rv is self, "Popped wrong request context. (%r instead of %r)" % (
                rv,
                self,
            )

    def auto_pop(self, exc):
        if self.request.environ.get("flask._preserve_context") or (
            exc is not None and self.app.preserve_context_on_exception
        ):
            self.preserved = True
            self._preserved_exc = exc
        else:
            self.pop(exc)

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        # do not pop the request stack if we are in debug mode and an
        # exception happened.  This will allow the debugger to still
        # access the request object in the interactive shell.  Furthermore
        # the context can be force kept alive for the test client.
        # See flask.testing for how this works.
        self.auto_pop(exc_value)

        if BROKEN_PYPY_CTXMGR_EXIT and exc_type is not None:
            reraise(exc_type, exc_value, tb)

    def __repr__(self):
        return "<%s '%s' [%s] of %s>" % (
            self.__class__.__name__,
            self.request.url,
            self.request.method,
            self.app.name,
        )



from ._MessageProcessorExecutorFactory import MessageProcessorExecutorFactory
from ._DefaultMessageProcessorExecutor import DefaultMessageProcessorExecutor
from ._MessageProcessorCallback import MessageProcessorCallback


class DefaultMessageProcessorExecutorFactory(MessageProcessorExecutorFactory):
    def __init__(self, callbacks):
        # type: (List[MessageProcessorCallback]) -> None
        self.callbacks = callbacks

    def build_executor(self, message_processor):
        return DefaultMessageProcessorExecutor(message_processor, self.callbacks)



from production.command import CommandType
from production.model.request import UndoRequest

from ._CommandProcessor import CommandProcessor


class UndoProcessor(CommandProcessor):
    def get_command_type(self):
        return CommandType.undo

    def parse_data(self, data):
        return UndoRequest(data.split('\0')[0])



from mebuhi.event.impl import EventManager, IdleCleaner
from mebuhi.http import FlaskAdapter
from mebuhi.http.processor import SendMessageProcessor, StartListenProcessor, ListenProcessor
from messagebus import TokenCreator, TokenPriority
from messagehandler import MessageHandlerBuilder
from simplehttprouter import Route

from ._MebuhiMessageHandler import MebuhiMessageHandler

PERIPHERALS_GROUP = "6"
TK_PRN_PRINT = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "101")


class MebuhiMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, message_bus, idle_timeout, bump_timeout, cleaner_interval, host, port, logger):
        self.message_bus = message_bus
        self.idle_timeout = idle_timeout
        self.bump_timeout = bump_timeout
        self.cleaner_interval = cleaner_interval
        self.host = host
        self.port = port
        self.logger = logger
        self.message_handler = None
        self.flask_adapter = None
        self.idle_cleaner = None

    def build_singletons(self):
        event_manager = EventManager(self.message_bus, self.idle_timeout, self.bump_timeout, self.logger)
        send_message_processor = SendMessageProcessor(self.message_bus, self.logger)
        start_listen_processor = StartListenProcessor(event_manager, self.logger)
        listen_processor = ListenProcessor(event_manager, self.logger)
        self.flask_adapter = FlaskAdapter({
            Route("GET", "/mwapp/events/listen"): listen_processor.handle,
            Route("GET", "/mwapp/events/start"): start_listen_processor.handle,
            Route("POST", "/mwapp/services/{serviceType}/{serviceName}"): send_message_processor.handle,
            Route("GET", "/mwapp/services/{serviceType}/{serviceName}/"): send_message_processor.handle
        }, self.logger)
        self.flask_adapter.start("MebuhiFlaskApp", self.host, self.port)

        self.idle_cleaner = IdleCleaner(event_manager, self.cleaner_interval, self.logger)
        self.idle_cleaner.start()

        self.message_handler = MebuhiMessageHandler(event_manager)

    def build_message_handler(self):
        return self.message_handler

    def destroy_message_handler(self, message_handler):
        return

    def destroy_singletons(self):
        self.idle_cleaner.stop()
        self.flask_adapter.stop()



from datetime import datetime
from typing import List
from _OrderTender import OrderTender


class OrderSate(object):
    VOIDED = 4
    PAID = 5


class Order(object):
    def __init__(self, order_id, date, state, total, tenders):
        # type: (int, datetime, int, float, List[OrderTender]) -> None
        self.order_id = order_id
        self.date = date
        self.state = state
        self.total = total
        self.tenders = tenders



from logging import getLogger

from nperipherals_core.model.device_status import DeviceStatus
from nperipherals_core.model.repository import NotificationRepository
from nperipherals_core.service import ReaderDeviceServiceThread
from typing import Optional

from externals.device import NetworkDevice
from nperipherals_core.model.custom_exceptions import DeviceNotConnected
from nperipherals_core.model.device import ReaderDevice, DeviceServiceThread

logger = getLogger(__name__)


class NetworkScannerDevice(ReaderDevice, NetworkDevice):
    def __init__(self, device_configurations):
        super(NetworkScannerDevice, self).__init__(logger, device_configurations)

    def _create_service_thread(self, repository):
        # type: (NotificationRepository) -> Optional[DeviceServiceThread]
        return ReaderDeviceServiceThread(self, repository)

    def read(self):
        # type: () -> Optional[str]

        try:
            data = self.socket.recv(self.SOCKET_BUF_SIZE)
            if data:
                self.logger.debug("Device [{}] got data [{}]".format(self.get_name(), data))
                return data

        except Exception as e:
            self.logger.exception("Error getting socket data with error [{}]".format(repr(e)))

        raise DeviceNotConnected()

    def get_status(self):
        return DeviceStatus.Ok



from nperipherals_core.model.device import Device


class PrinterStatusInteractor(object):
    def __init__(self, writer_device):
        # type: (Device) -> None

        self.device = writer_device

    def get_device_status(self):
        connected = False
        try:
            self.device.initialize()
            self.device.connect()
            connected = True
            return self.device.get_status()
        finally:
            if connected:
                self.device.disconnect()



from logging import Logger
from datetime import datetime, timedelta

from production.box import ProductionBox
from production.model import Item, ProductionOrder
from typing import List, Dict, Union, Any, Optional

from ._OrderChangerProductionBox import OrderChangerProductionBox


class CourseBox(OrderChangerProductionBox):
    def __init__(self, name, sons, courses, publish_scheduler, logger=None):
        # type: (str, Union[ProductionBox, List[ProductionBox]], List[CourseConfiguration], Any, Logger) -> None  # noqa

        super(CourseBox, self).__init__(name, sons, logger)
        self.courses = courses
        self.publish_scheduler = publish_scheduler
        self.course_dict = {}
        self.course_name_dict = {}
        for course_index in range(0, len(self.courses)):
            for part_code in courses[course_index].products:
                self.course_dict[part_code] = course_index
                self.course_name_dict[courses[course_index].name] = course_index

    def change_order(self, order):
        # type: (ProductionOrder) -> Union[ProductionOrder, List[ProductionOrder]]

        if len(self.courses) == 0 or order.skip_course:
            return order

        courses = self.separate_product_in_courses(order)
        valid_courses_count = self.count_valid_courses(courses)
        if valid_courses_count <= 1:
            return order

        orders = []
        split_index = 0
        all_previous_is_ready = True
        current_course = self.get_first_not_ready_course(courses)

        for course in courses:
            if len(course.products) == 0:
                continue

            if course == current_course:
                current_order = self._get_current_course_order(courses, current_course, order)
                new_order = current_order.clone(index=split_index)
            else:
                product = course.products if all_previous_is_ready else []
                new_order = order.clone(product, split_index)
            orders.append(new_order)
            split_index += 1

            all_previous_is_ready = course.ready and all_previous_is_ready

        return orders

    def _get_current_course_order(self, courses, current_course, order):
        # type: (List[Course], Course, ProductionOrder) -> ProductionOrder

        now = datetime.now()
        last_done_time = self.get_last_done_time_of_course(courses[current_course.index - 1])
        course_delay_ok = self.has_elapsed_course_interval(now, courses[current_course.index - 1], last_done_time)
        if last_done_time is None or course_delay_ok:
            order.items = current_course.products

            if last_done_time is not None:
                display_time = self.get_display_time(courses[current_course.index - 1], last_done_time)
                order.display_time = display_time.isoformat()
        else:
            order.items = []

            wait_time = self.calculate_wait_time(now, courses[current_course.index - 1], last_done_time)
            if wait_time > 0:
                self.publish_scheduler.schedule_publish(order.order_id, wait_time)

        return order

    @staticmethod
    def get_first_not_ready_course(courses):
        # type: (List[Course]) -> Optional[Course]

        for course in courses:
            if not course.ready:
                return course
        return None

    @staticmethod
    def count_valid_courses(courses):
        # type: (List[Course]) -> int

        valid_courses = [x for x in courses if len(x.products) > 0]
        return len(valid_courses)

    def separate_product_in_courses(self, order):
        # type: (ProductionOrder) -> List

        course_products = {k: [] for k in range(-1, len(self.courses))}
        course_ready = {k: True for k in range(-1, len(self.courses))}
        course_done = {k: True for k in range(-1, len(self.courses))}

        self._populate_course_products(course_products, order.items)
        self._handle_products_without_course(course_products, True)
        for course_index in range(0, len(self.courses)):
            for item in course_products[course_index]:
                if item.multiplied_qty > 0 and not item.is_prepared():
                    course_ready[course_index] = False

                if item.multiplied_qty > 0 and not item.is_done():
                    course_done[course_index] = False

        self._send_fire_items_to_first_available_course(course_products)

        return self._create_courses(course_products, course_ready, course_done)

    @staticmethod
    def _create_courses(course_products, course_ready, course_done):
        # type: (Dict[int, List[Item]], Dict, Dict) -> List[Course]

        courses = []
        for index in range(0, len(course_products)):
            products = course_products[index]
            course = CourseBox.Course(products, course_ready[index], course_done[index], index)
            courses.append(course)

        for index in range(0, len(courses)):
            all_course_items_canceled = True
            for product in courses[index].products:
                if not product.ignore_removed_item():
                    all_course_items_canceled = False
                    break

            if all_course_items_canceled and index < len(courses) - 1:
                for product in courses[index].products:
                    courses[index + 1].products.append(product)

        return courses

    def _populate_course_products(self, course_products, items):
        # type: (Dict[int, List[Item]], List[Item]) -> None

        for item in items:
            self._populate_item_course(item, course_products)

        self._handle_products_without_course(course_products)

    def _send_fire_items_to_first_available_course(self, course_products):
        # type: (Dict[int, List[Item]]) -> None

        course_qty, first_course_with_products = self._count_courses(course_products)
        self._move_fire_products(course_products, course_qty, first_course_with_products)

    @staticmethod
    def _move_fire_products(course_products, course_qty, first_course_with_products):
        # type: (Dict[int, List[Item]], int, int) -> None

        if course_qty > 1:
            for index in range(first_course_with_products + 1, len(course_products)):
                for item in course_products[index][:]:
                    if "fire" in item.properties or "fire" in item.tags:
                        course_products[first_course_with_products].append(item)
                        course_products[index].remove(item)

    @staticmethod
    def _count_courses(course_products):
        # type: (Dict[int, List[Item]]) -> (int, int)

        course_qty = 0
        first_course_with_products = None
        for course_index in course_products:
            if len(course_products[course_index]) > 0:
                course_qty += 1
                if first_course_with_products is None:
                    first_course_with_products = course_index
        return course_qty, first_course_with_products

    def _populate_item_course(self, item, course_products):
        # type: (Item, Dict[int, List[Item]]) -> None

        if self.item_has_course_set(item):
            item_course_index = self.get_course_index_from_set_course(item)
            course_products[item_course_index].append(item)
            return

        if item.part_code in self.course_dict:
            item_course_index = self.course_dict[item.part_code]
            course_products[item_course_index].append(item)
            return

        if item.item_type not in ("COMBO", "OPTION"):
            course_products[-1].append(item)
            return

        son_course_products = {k: [] for k in range(-1, len(self.courses))}
        self._populate_course_products(son_course_products, item.items)

        for key, value in son_course_products.items():
            if len(value) == 0:
                continue

            new_combo = item.clone()
            new_combo.items = value
            course_products[key].append(new_combo)

    def _handle_products_without_course(self, course_products, process_products_without_course=False):
        has_product_without_course = -1 in course_products

        if has_product_without_course:
            for course_index in range(0, len(self.courses)):
                if len(course_products[course_index]) > 0:
                    course_products[course_index].extend(course_products[-1])
                    del course_products[-1]
                    break
            else:
                if process_products_without_course:
                    course_products[0].extend(course_products[-1])
                    del course_products[-1]

    def item_has_course_set(self, item):
        # type: (Item) -> bool

        if item.properties is not None:
            return "course" in item.properties and item.properties["course"] in self.course_name_dict

        return False

    def get_course_index_from_set_course(self, item):
        # type: (Item) -> int

        return self.course_name_dict[item.properties["course"]]

    def has_elapsed_course_interval(self, now, course, last_done_time):
        # type: (datetime, Course, datetime) -> bool
        if last_done_time is None:
            return True

        next_course_ready = last_done_time + timedelta(seconds=self.courses[course.index].next_course_wait_time)
        return int((now - next_course_ready).total_seconds()) >= 0

    def get_display_time(self, course, last_done_time):
        # type: (Course, datetime) -> datetime

        course_wait = self.courses[course.index].next_course_wait_time
        return last_done_time + timedelta(seconds=course_wait)

    def calculate_wait_time(self, now, course, last_done_time):
        # type: (datetime, Course, datetime) -> int

        wait_time = ((last_done_time + timedelta(seconds=self.courses[course.index].next_course_wait_time)) - now)
        return int(wait_time.total_seconds())

    @staticmethod
    def get_last_done_time_of_course(course):
        # type: (Course) -> Optional[datetime]

        last_done_time = None
        for item in course.products:
            done_time = item.get_last_prepared_time()
            if done_time is not None and (last_done_time is None or last_done_time < done_time):
                last_done_time = done_time

        return last_done_time

    class CourseConfiguration(object):
        def __init__(self, name, products, next_course_wait_time):
            self.name = name
            self.products = products
            self.next_course_wait_time = next_course_wait_time

    class Course(object):
        def __init__(self, products, ready, done, index):
            self.products = products
            self.ready = ready
            self.done = done
            self.index = index



"""External interface to the BeautifulSoup HTML parser.
"""

__all__ = ["fromstring", "parse", "convert_tree"]

import re
from lxml import etree, html

try:
    from bs4 import (
        BeautifulSoup, Tag, Comment, ProcessingInstruction, NavigableString,
        Declaration, Doctype)
    _DECLARATION_OR_DOCTYPE = (Declaration, Doctype)
except ImportError:
    from BeautifulSoup import (
        BeautifulSoup, Tag, Comment, ProcessingInstruction, NavigableString,
        Declaration)
    _DECLARATION_OR_DOCTYPE = Declaration


def fromstring(data, beautifulsoup=None, makeelement=None, **bsargs):
    """Parse a string of HTML data into an Element tree using the
    BeautifulSoup parser.

    Returns the root ``<html>`` Element of the tree.

    You can pass a different BeautifulSoup parser through the
    `beautifulsoup` keyword, and a diffent Element factory function
    through the `makeelement` keyword.  By default, the standard
    ``BeautifulSoup`` class and the default factory of `lxml.html` are
    used.
    """
    return _parse(data, beautifulsoup, makeelement, **bsargs)


def parse(file, beautifulsoup=None, makeelement=None, **bsargs):
    """Parse a file into an ElemenTree using the BeautifulSoup parser.

    You can pass a different BeautifulSoup parser through the
    `beautifulsoup` keyword, and a diffent Element factory function
    through the `makeelement` keyword.  By default, the standard
    ``BeautifulSoup`` class and the default factory of `lxml.html` are
    used.
    """
    if not hasattr(file, 'read'):
        file = open(file)
    root = _parse(file, beautifulsoup, makeelement, **bsargs)
    return etree.ElementTree(root)


def convert_tree(beautiful_soup_tree, makeelement=None):
    """Convert a BeautifulSoup tree to a list of Element trees.

    Returns a list instead of a single root Element to support
    HTML-like soup with more than one root element.

    You can pass a different Element factory through the `makeelement`
    keyword.
    """
    root = _convert_tree(beautiful_soup_tree, makeelement)
    children = root.getchildren()
    for child in children:
        root.remove(child)
    return children


# helpers

def _parse(source, beautifulsoup, makeelement, **bsargs):
    if beautifulsoup is None:
        beautifulsoup = BeautifulSoup
    if hasattr(beautifulsoup, "HTML_ENTITIES"):  # bs3
        if 'convertEntities' not in bsargs:
            bsargs['convertEntities'] = 'html'
    if hasattr(beautifulsoup, "DEFAULT_BUILDER_FEATURES"):  # bs4
        if 'features' not in bsargs:
            bsargs['features'] = 'html.parser'  # use Python html parser
    tree = beautifulsoup(source, **bsargs)
    root = _convert_tree(tree, makeelement)
    # from ET: wrap the document in a html root element, if necessary
    if len(root) == 1 and root[0].tag == "html":
        return root[0]
    root.tag = "html"
    return root


_parse_doctype_declaration = re.compile(
    r'(?:\s|[<!])*DOCTYPE\s*HTML'
    r'(?:\s+PUBLIC)?(?:\s+(\'[^\']*\'|"[^"]*"))?'
    r'(?:\s+(\'[^\']*\'|"[^"]*"))?',
    re.IGNORECASE).match


class _PseudoTag:
    # Minimal imitation of BeautifulSoup.Tag
    def __init__(self, contents):
        self.name = 'html'
        self.attrs = []
        self.contents = contents

    def __iter__(self):
        return self.contents.__iter__()


def _convert_tree(beautiful_soup_tree, makeelement):
    if makeelement is None:
        makeelement = html.html_parser.makeelement

    # Split the tree into three parts:
    # i) everything before the root element: document type
    # declaration, comments, processing instructions, whitespace
    # ii) the root(s),
    # iii) everything after the root: comments, processing
    # instructions, whitespace
    first_element_idx = last_element_idx = None
    html_root = declaration = None
    for i, e in enumerate(beautiful_soup_tree):
        if isinstance(e, Tag):
            if first_element_idx is None:
                first_element_idx = i
            last_element_idx = i
            if html_root is None and e.name and e.name.lower() == 'html':
                html_root = e
        elif declaration is None and isinstance(e, _DECLARATION_OR_DOCTYPE):
            declaration = e

    # For a nice, well-formatted document, the variable roots below is
    # a list consisting of a single <html> element. However, the document
    # may be a soup like '<meta><head><title>Hello</head><body>Hi
    # all<\p>'. In this example roots is a list containing meta, head
    # and body elements.
    if first_element_idx is None:
        pre_root = post_root = []
        roots = beautiful_soup_tree.contents
    else:
        pre_root = beautiful_soup_tree.contents[:first_element_idx]
        roots = beautiful_soup_tree.contents[first_element_idx:last_element_idx+1]
        post_root = beautiful_soup_tree.contents[last_element_idx+1:]

    # Reorganize so that there is one <html> root...
    if html_root is not None:
        # ... use existing one if possible, ...
        i = roots.index(html_root)
        html_root.contents = roots[:i] + html_root.contents + roots[i+1:]
    else:
        # ... otherwise create a new one.
        html_root = _PseudoTag(roots)

    convert_node = _init_node_converters(makeelement)

    # Process pre_root
    res_root = convert_node(html_root)
    prev = res_root
    for e in reversed(pre_root):
        converted = convert_node(e)
        if converted is not None:
            prev.addprevious(converted)
            prev = converted

    # ditto for post_root
    prev = res_root
    for e in post_root:
        converted = convert_node(e)
        if converted is not None:
            prev.addnext(converted)
            prev = converted

    if declaration is not None:
        try:
            # bs4 provides full Doctype string
            doctype_string = declaration.output_ready()
        except AttributeError:
            doctype_string = declaration.string

        match = _parse_doctype_declaration(doctype_string)
        if not match:
            # Something is wrong if we end up in here. Since soupparser should
            # tolerate errors, do not raise Exception, just let it pass.
            pass
        else:
            external_id, sys_uri = match.groups()
            docinfo = res_root.getroottree().docinfo
            # strip quotes and update DOCTYPE values (any of None, '', '...')
            docinfo.public_id = external_id and external_id[1:-1]
            docinfo.system_url = sys_uri and sys_uri[1:-1]

    return res_root


def _init_node_converters(makeelement):
    converters = {}
    ordered_node_types = []

    def converter(*types):
        def add(handler):
            for t in types:
                converters[t] = handler
                ordered_node_types.append(t)
            return handler
        return add

    def find_best_converter(node):
        for t in ordered_node_types:
            if isinstance(node, t):
                return converters[t]
        return None

    def convert_node(bs_node, parent=None):
        # duplicated in convert_tag() below
        try:
            handler = converters[type(bs_node)]
        except KeyError:
            handler = converters[type(bs_node)] = find_best_converter(bs_node)
        if handler is None:
            return None
        return handler(bs_node, parent)

    def map_attrs(bs_attrs):
        if isinstance(bs_attrs, dict):  # bs4
            attribs = {}
            for k, v in bs_attrs.items():
                if isinstance(v, list):
                    v = " ".join(v)
                attribs[k] = unescape(v)
        else:
            attribs = dict((k, unescape(v)) for k, v in bs_attrs)
        return attribs

    def append_text(parent, text):
        if len(parent) == 0:
            parent.text = (parent.text or '') + text
        else:
            parent[-1].tail = (parent[-1].tail or '') + text

    # converters are tried in order of their definition

    @converter(Tag, _PseudoTag)
    def convert_tag(bs_node, parent):
        attrs = bs_node.attrs
        if parent is not None:
            attribs = map_attrs(attrs) if attrs else None
            res = etree.SubElement(parent, bs_node.name, attrib=attribs)
        else:
            attribs = map_attrs(attrs) if attrs else {}
            res = makeelement(bs_node.name, attrib=attribs)

        for child in bs_node:
            # avoid double recursion by inlining convert_node(), see above
            try:
                handler = converters[type(child)]
            except KeyError:
                pass
            else:
                if handler is not None:
                    handler(child, res)
                continue
            convert_node(child, res)
        return res

    @converter(Comment)
    def convert_comment(bs_node, parent):
        res = html.HtmlComment(bs_node)
        if parent is not None:
            parent.append(res)
        return res

    @converter(ProcessingInstruction)
    def convert_pi(bs_node, parent):
        if bs_node.endswith('?'):
            # The PI is of XML style (<?as df?>) but BeautifulSoup
            # interpreted it as being SGML style (<?as df>). Fix.
            bs_node = bs_node[:-1]
        res = etree.ProcessingInstruction(*bs_node.split(' ', 1))
        if parent is not None:
            parent.append(res)
        return res

    @converter(NavigableString)
    def convert_text(bs_node, parent):
        if parent is not None:
            append_text(parent, unescape(bs_node))
        return None

    return convert_node


# copied from ET's ElementSoup

try:
    from html.entities import name2codepoint  # Python 3
except ImportError:
    from htmlentitydefs import name2codepoint


handle_entities = re.compile(r"&(\w+);").sub


try:
    unichr
except NameError:
    # Python 3
    unichr = chr


def unescape(string):
    if not string:
        return ''
    # work around oddities in BeautifulSoup's entity handling
    def unescape_entity(m):
        try:
            return unichr(name2codepoint[m.group(1)])
        except KeyError:
            return m.group(0)  # use as is
    return handle_entities(unescape_entity, string)



from logging import getLogger
from messagebus import DataType, DefaultToken, Message
from messageprocessorutil import UuidMessageProcessor
from nperipherals_core.interactor.cash_drawer_interactor import CashDrawerOpenInteractor

logger = getLogger(__name__)


class CashDrawerOpenProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (CashDrawerOpenInteractor) -> None
        super(CashDrawerOpenProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # we don't expect data
        pass

    def call_business(self, obj):
        return self.interactor.open_drawer()

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, str(result)))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.string, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.string, str(exception)))



# encoding: utf-8
import datetime

from old_helper import BaseRepository
from ..model import Order


class OrderRepository(BaseRepository):

    def __init__(self, mbcontext, pos_list):
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_sales_by_date(self, user_id, start_date, end_date):
        def inner_func(conn):
            start_date_formatted = datetime.datetime.strftime(start_date, "%Y-%m-%dT%H:%M:%S")
            end_date_formatted = datetime.datetime.strftime(end_date, "%Y-%m-%dT%H:%M:%S")

            query = self._OrdersQuery.format(user_id, start_date_formatted, end_date_formatted)

            orders = []
            for x in conn.select(query):
                orders.append(Order(x.get_entry("Timestamp"), x.get_entry("SaleType"), float(x.get_entry("Amount"))))
            return orders

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    _OrdersQuery = """\
    SELECT ocp.Value as Timestamp, st.TypeDescr as SaleType, o.TotalGross as Amount FROM Orders o
    INNER JOIN OrderCustomProperties ocp ON o.OrderId = ocp.OrderID
    INNER JOIN SaleType st ON o.SaleType = st.TypeId
    WHERE  o.stateid = 5 
        AND o.SessionId like '%user={0},%' 
        AND KEY = 'FISCALIZATION_DATE'
        AND ocp.value >= '{1}'
        AND ocp.value < '{2}'
    ORDER BY Timestamp
    """



import os
import sys
import threading
import json
import sqlite3
import datetime
import Queue
from threading import Lock, Condition
from xml.etree import cElementTree as eTree

import msgbus
import systools
import syserrors
from application.configurator import Configurator

from mwapp.components import BaseComponent
from typing import Callable


class AuditEventConsumer(BaseComponent):
    def __init__(self, configs, process):
        # type: (Configurator, Callable) -> None

        service_name = configs.service_name
        service_type = configs.service_type
        required_services = configs.required_services

        super(AuditEventConsumer, self).__init__(service_name, service_type, required_services)
        self.logger = configs.logger

        self.service_name = service_name
        self.service_type = service_type
        self.eventFilter = []
        self.eventFilterData = {}
        self.last_row_id = None
        self.initial_period = None
        self.last_period = None
        self.queue_full = False
        self.lock = Lock()
        self.condition = Condition()
        self.terminating = False
        self.skip_event = False
        self.mode = 'DEFAULT'

        self.mbcontext.MB_EasyEvtSubscribe('AUDITLOGGER')
        self.mbcontext.MB_EasyEvtSubscribe(self.name)
        self._read_loader(configs)

        self.q = Queue.Queue()
        self.buffer_q = Queue.Queue(maxsize=self.queue_max_size)

        self.ctrl_conn = None
        self.process = process
        with self.lock:
            self.init_control_file()

    def _read_loader(self, configs):
        group = self.service_name + ".{}"
        cfg_reader = configs.cfg_reader
        self.initial_period = int(cfg_reader.find_value(group.format("initialPeriod"), "0"))
        if self.initial_period == 0:
            self.logger.error("Configuration error: initialPeriod is undefined")
            sys.exit(msgbus.SE_NOTFOUND)

        control = cfg_reader.find_value(group.format("controlFilePath"))
        self.controlFilePath = systools.sys_parsefilepath(control)
        control_path = os.path.dirname(self.controlFilePath)
        if not os.path.exists(control_path):
            os.mkdir(control_path, 0777)
        self.queue_max_size = int(cfg_reader.find_value(group.format("queueMaxSize"), "100"))
        self.resync_required = cfg_reader.find_value(group.format("resyncRequired"), "true").lower() == "true"

        for evt_filter in cfg_reader.find_values(group.format("eventFilter")) or []:
            event = evt_filter.split('/')
            self.eventFilter.append(event[0])
            if len(event) > 1:
                self.eventFilterData[event[0]] = event[1:]
        self.evt_batch_size = int(cfg_reader.find_value(group.format("eventBatchSize"), "0"))
        self.evt_batch_time = cfg_reader.find_value(group.format("eventBatchTime"), "false").lower() == "true"

        if self.evt_batch_size:
            self.mode = 'BATCH_SIZE'
        if self.evt_batch_time:
            self.mode = 'BATCH_TIME'

    def _create_db_conn(self):
        ctrl_conn = None
        try:
            ctrl_conn = sqlite3.connect(self.controlFilePath)
            ctrl_conn.isolation_level = None
        except Exception as _:  # noqa
            self.logger.exception("Failed to get the thread's control file handle.")

        return ctrl_conn

    def init_control_file(self):
        self.ctrl_conn = self._create_db_conn()
        try:
            self._create_events_sent_table(self.ctrl_conn)

            c = self.ctrl_conn.cursor()
            r = c.execute("""
                SELECT Period,LastRowId 
                    FROM EventsSent 
                    WHERE period >= (SELECT max(Period) FROM EventsSent) order by LastRowId desc limit 1;
            """).fetchone()
            if r:
                self.last_period = r[0]
                self.last_row_id = r[1]
            else:
                self.last_period = self.initial_period
                self.last_row_id = "-1"
            c.close()
        except Exception as _:  # noqa
            self.logger.exception("Failure during control file initialization.")
            sys.exit(msgbus.SE_BADPARAM)

    def _create_events_sent_table(self, ctrl_conn):
        if not ctrl_conn:
            ctrl_conn = self.ctrl_conn

        ctrl_conn.execute("""
                CREATE TABLE IF NOT EXISTS EventsSent (
                    Period      INTEGER NOT NULL,
                    LastRowId   INTEGER NOT NULL,
                    PRIMARY KEY (LastRowId)
                );
            """)
        ctrl_conn.execute("CREATE INDEX IF NOT EXISTS `index_LastRowId` ON `EventsSent` (`LastRowId` DESC);")

    def resync(self, last_period, last_row_id):
        self.logger.info("Running resync.")
        try:
            evt_types = (",".join(self.eventFilter))
            params = [last_period, evt_types, last_row_id]
            d = "\0".join(str(p) for p in params)
            self.mbcontext.MB_EasySendMessage("AuditLogger", msgbus.TK_AUDITLOG_RESYNC, format=msgbus.FM_PARAM, data=d)
        except Exception as _:  # noqa
            if self.resync_required:
                self.logger.exception("Exception occurred during resynchronization.")
                sys.exit(syserrors.SE_MB_ERR)

    def force_new_period(self, new_period, ctrl_conn=None):
        if not ctrl_conn:
            ctrl_conn = self.ctrl_conn

        new_period_date = datetime.datetime.strptime(str(new_period), '%Y%m%d')
        last_period_date = datetime.datetime.strptime(str(self.last_period), '%Y%m%d')

        log_message = "Reprocess Sales Command Date {} - Last Period Sent {}"
        self.logger.info(log_message.format(new_period_date, last_period_date))

        if new_period_date <= last_period_date:
            self.lock.acquire()
            ctrl_conn.execute("DELETE FROM EventsSent WHERE Period >= %s" % new_period)
            self._harakiri()

    def start(self):
        t = threading.Thread(target=self.worker)
        t.daemon = True
        t.start()

        while not self.terminating:
            msg = self.mbcontext.MB_EasyGetMessage()
            if not msg or msg.token == msgbus.TK_CMP_TERM_NOW:
                self._send_nack(msg)
                self.terminating = True
                with self.condition:
                    self.condition.notify()
                break

            elif msg.token == msgbus.TK_EVT_EVENT:
                params = msg.data.split('\0') if msg.data else []

                if not params:
                    self._send_nack(msg)
                    continue

                xml, subject, event_type = params[0], params[1], params[2]
                if subject == "SKIPEVENT":
                    self.skip_event = True
                    self._send_ack(msg)
                    continue

                if subject == "AUDITLOGGER":
                    if event_type not in self.eventFilter or self.queue_full:
                        self._send_ack(msg)
                        continue

                    item = [xml, event_type]
                    self._insert_event_into_queue(item)

                elif subject == "BohPump":
                    if event_type == 'SendEvents':
                        item_list = []
                        for _ in range(self.buffer_q.qsize()):
                            item = self.buffer_q.get()
                            item_list.append(item)
                        self.q.put(item_list, False)

                else:
                    self._send_nack(msg)

            self._send_nack(msg)

        if self.ctrl_conn:
            self.ctrl_conn.close()

    def _insert_event_into_queue(self, item):
        try:
            if not self.evt_batch_size and not self.evt_batch_time:
                item = [item]
                self.q.put(item, False)
            else:
                self.buffer_q.put(item, False)

            if self.evt_batch_size and self.buffer_q.qsize() >= self.evt_batch_size:
                item_list = []
                for _ in range(self.evt_batch_size):
                    item = self.buffer_q.get()
                    item_list.append(item)
                self.q.put(item_list, False)
        except Queue.Full:
            self.logger.error("Queue Full - Locking: Size {}".format(self.q.qsize()))
            self.queue_full = True

    @staticmethod
    def fib(n_sequence):
        a, b = 1, 1
        for _ in range(n_sequence - 1):
            a, b = b, a + b
        return a

    def worker(self):
        self.logger.info("Starting worker process...")
        self.resync(self.last_period, self.last_row_id)

        ctrl_conn = self._create_db_conn()
        while not self.terminating:
            last_error = None
            retry_count = 1
            item = self._get_new_item(ctrl_conn)

            while last_error != msgbus.SE_SUCCESS and not self.terminating:
                try:
                    last_error = self.process(item)
                    if last_error != syserrors.SE_SUCCESS and not self.skip_event:
                        if retry_count < 14:
                            retry_count += 1

                        self.logger.info("Error sending event, retry in {} seconds ...".format(self.fib(retry_count)))
                        with self.condition:
                            self.condition.wait(self.fib(retry_count))
                        continue

                    if self.skip_event:
                        self.logger.info("Event id {} skipped".format(self.last_row_id))
                        last_error = msgbus.SE_SUCCESS
                        self.skip_event = False

                    evt_period = self._mark_events_as_sent(item, ctrl_conn)
                    self._clean_control_file(evt_period, ctrl_conn)

                    if self.queue_full and self.buffer_q.qsize() == 0 and self.q.qsize() == 0:
                        self._harakiri()

                except Exception as _:  # noqa
                    self.logger.exception("Could not start worker thread")

        self.logger.info("Finishing worker process")

    def _clean_control_file(self, period, ctrl_conn):
        if period is None:
            return

        evt_period = datetime.datetime.strptime(str(period), '%Y%m%d')
        last_period = datetime.datetime.strptime(str(self.last_period), '%Y%m%d')

        if evt_period > last_period:
            date_subtract = last_period - datetime.timedelta(days=15)
            ctrl_conn.execute("DELETE FROM EventsSent WHERE Period <= %s" % int(date_subtract.strftime('%Y%m%d')))
            self.last_period = period

    def _mark_events_as_sent(self, item, ctrl_conn):
        evt_period = None
        for idx, q_item in enumerate(item):
            event = eTree.XML(q_item[0])
            audit_log = event.find("AuditLog")
            evt_row_id = audit_log.get("rowId")
            evt_period = audit_log.get("period")
            if not self.lock.locked():
                with self.lock:
                    query = "INSERT OR IGNORE INTO EventsSent(Period, LastRowId) VALUES(:period, :rowid)"
                    ctrl_conn.execute(query, {"period": evt_period, "rowid": evt_row_id})

            self.last_row_id = evt_row_id
        return evt_period

    def _harakiri(self):
        self.logger.info("Exiting to force resync events processed ...")
        hv_service = self.mbcontext.hv_service
        self.mbcontext.MB_SendOneWayMessage(hv_service, msgbus.TK_HV_COMPSTOP, msgbus.FM_STRING, "BohPump")

    def _get_new_item(self, ctrl_conn):
        item = self.q.get()
        for idx, q_item in enumerate(item):
            event = eTree.XML(q_item[0])
            audit_log = event.find("AuditLog")
            evt_row_id = audit_log.get("rowId")
            evt_period = audit_log.get("period")

            self.logger.info("Processing event {} for period {} ...".format(evt_row_id, evt_period))
            if int(evt_row_id) <= int(self.last_row_id):
                query = """SELECT 1 FROM EventsSent WHERE LastRowId = (:last_row_id) and Period = (:period);"""
                evt_check = ctrl_conn.execute(query, {"last_row_id": evt_row_id, "period": evt_period}).fetchone()
                if evt_check:
                    item.pop(idx)
        return item

    def send_response(self, msg, code=200):
        msg.token = msgbus.TK_SYS_ACK if 200 < code < 300 else msgbus.TK_SYS_NAK
        self.mbcontext.MB_ReplyMessage(msg, format=msgbus.FM_PARAM, data=json.dumps({'code': code}))

    def renew_service(self):
        self._harakiri()



from escposformatter import DefaultEscPosFormatter
from htmlformatter import HtmlFormatter
from report import ReportParser
from sysactions import generate_report, print_text, show_print_preview, show_messagebox

from enum import Enum


class PrinterType(Enum):
    esc_pos = "ESC_POS"
    apos = "APOS"


def print_report(pos_id, report_name, preview, *args):
    report_data = generate_report(report_name, pos_id, *args)

    report = ReportParser().parse(report_data)
    if report is not None:
        if preview:
            print_preview_report = HtmlFormatter().format(report)
            if show_print_preview(pos_id, print_preview_report.encode("utf-8")) in (1, None):
                return False

        text = DefaultEscPosFormatter().format(report)
        print_text(pos_id, None, text)
    elif report is None:
        show_messagebox(pos_id, "$NO_REPORT_DATA", "$ERROR")
    else:
        return print_text(pos_id, None, report_data, preview)



class Action(object):
    def __init__(self, obj):
        self.order = int(obj.get("order")) if obj.get("order") else None
        self.name = obj.get("name")
        self.value = obj.get("value")
        self.times_to_apply = None



import datetime  # noqa

from _Transfer import Transfer  # noqa


class TransfersByDayDto(object):
    def __init__(self, day, transfers_by_date):
        # type:(datetime, [Transfer]) -> None
        self.day = day
        self.transfers_by_date = transfers_by_date



import logging
from typing import Dict, Any
from datetime import datetime
from dateutil import tz
from xml.etree import cElementTree as eTree

from pos_model import Order
import sysactions

from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys
from comp_exceptions import FiscalGetIdException

logger = logging.getLogger("FiscalWrapper")


class NfceEmissionType(object):
    normal = 1
    contingency = 9


class IdeBuilder(NfceXmlPartBuilder):

    def __init__(self,
                 serie,
                 initial_order_id,
                 c_uf,
                 mod,
                 c_mun_fg,
                 ambiente, versao_ws):
        #  type: (int, int, int, int, int, int, int) -> None
        self.serie = serie
        self.initial_order_id = initial_order_id
        self.c_uf = c_uf
        self.mod = mod
        self.c_mun_fg = c_mun_fg
        self.ambiente = ambiente
        self.chave = None
        self.versao_ws = versao_ws

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        is_in_contingency = context[ContextKeys.is_in_contingency]
        contingency_datetime = context[ContextKeys.contingency_datetime]
        contingency_reason = context[ContextKeys.contingency_reason]

        last_totaled_state = None
        for order_state in reversed(order.states):
            if order_state.name == "TOTALED":
                last_totaled_state = order_state
                break

        if last_totaled_state is None:
            raise Exception("The order was not totaled")

        data_emissao = self._format_date(last_totaled_state.timestamp)

        context[ContextKeys.data_emissao] = data_emissao
        serie_nota = int(self.serie)

        numero_nota = self._get_fiscal_id(order.pos_id, order.order_id)
        context[ContextKeys.fiscal_number] = numero_nota

        tp_emis = NfceEmissionType.contingency if is_in_contingency else NfceEmissionType.normal
        context[ContextKeys.emission_type] = tp_emis

        inner_xml = "<ide>" \
                    "<cUF>%s</cUF>" \
                    "<cNF>${CNF_RANDOM}</cNF>" \
                    "<natOp>venda</natOp>" % self.c_uf
        if self.versao_ws in (1, 3):
            inner_xml += "<indPag>0</indPag>"
        inner_xml += "<mod>%s</mod>" \
                     "<serie>%s</serie>" \
                     "<nNF>%s</nNF>" \
                     "<dhEmi>%s</dhEmi>" \
                     "<tpNF>1</tpNF>" \
                     "<idDest>1</idDest>" \
                     "<cMunFG>%s</cMunFG>" \
                     "<tpImp>4</tpImp>" \
                     "<tpEmis>%d</tpEmis>" \
                     "<cDV>${CDV}</cDV>" \
                     "<tpAmb>%s</tpAmb>" \
                     "<finNFe>1</finNFe>" \
                     "<indFinal>1</indFinal>" \
                     "<indPres>1</indPres>" \
                     "<procEmi>0</procEmi>" \
                     "<verProc>v1.00.00</verProc>" \
                     % (self.mod,
                        serie_nota,
                        numero_nota,
                        data_emissao,
                        self.c_mun_fg,
                        tp_emis,
                        self.ambiente)

        if is_in_contingency:
            inner_xml += "<dhCont>{}</dhCont><xJust>{}</xJust>".format(
                self._format_date(contingency_datetime),
                contingency_reason)

        inner_xml += "</ide>"

        return inner_xml

    @staticmethod
    def _get_fiscal_id(pos_id, order_id):
        # type: (int, int) -> int

        try:
            fiscal_id_property = "FISCAL_ID"

            model = sysactions.get_model(pos_id)
            pos_ot = sysactions.get_posot(model)

            xml = eTree.XML(pos_ot.getOrderCustomProperties(fiscal_id_property, orderid=str(order_id)))
            if xml:
                fiscal_id_node = xml.find('.//OrderProperty[@key="{}"]'.format(fiscal_id_property))
                if fiscal_id_node is not None:
                    fiscal_id = int(fiscal_id_node.get("value"))
                    return fiscal_id

        except Exception as _:
            logger.error("Error getting fiscal_id custom property from order [{}]".format(order_id))

        raise FiscalGetIdException()

    @staticmethod
    def _format_date(data):
        # type: (datetime) -> str
        local_zone = tz.tzlocal()
        data = data.replace(tzinfo=local_zone)
        data_str = data.strftime("%Y-%m-%dT%H:%M:%S%z")
        data_str = data_str[:22] + ":" + data_str[22:]
        return data_str
